"""Persisted queries support for GraphQL analytics.

Provides a simple file-based registry for persisted queries with
schema versioning to invalidate stale queries.
"""

from __future__ import annotations

import json
import logging
from dataclasses import dataclass
from pathlib import Path
from typing import Dict

from .errors import PersistedQueryError


logger = logging.getLogger(__name__)

# Current schema version - increment when schema changes break queries
SCHEMA_VERSION = "1.0"

# Path to the persisted queries registry file
_REGISTRY_PATH = Path(__file__).parent / "persisted_queries.json"


@dataclass
class PersistedQuery:
    """A persisted query entry."""

    id: str
    query: str
    schema_version: str
    description: str = ""


# In-memory cache of persisted queries
_QUERY_CACHE: Dict[str, PersistedQuery] = {}
_CACHE_LOADED = False


def _load_registry() -> None:
    """Load the persisted query registry from disk."""
    global _CACHE_LOADED

    if _CACHE_LOADED:
        return

    if not _REGISTRY_PATH.exists():
        logger.debug("No persisted queries registry found at %s", _REGISTRY_PATH)
        _CACHE_LOADED = True
        return

    try:
        with open(_REGISTRY_PATH, "r") as f:
            data = json.load(f)

        for entry in data.get("queries", []):
            query_id = entry.get("id")
            if query_id:
                _QUERY_CACHE[query_id] = PersistedQuery(
                    id=query_id,
                    query=entry.get("query", ""),
                    schema_version=entry.get("schema_version", SCHEMA_VERSION),
                    description=entry.get("description", ""),
                )

        logger.info("Loaded %d persisted queries", len(_QUERY_CACHE))
        _CACHE_LOADED = True

    except Exception as e:
        logger.warning("Failed to load persisted queries: %s", e)
        _CACHE_LOADED = True


def load_persisted_query(query_id: str) -> str:
    """
    Load a persisted query by ID.

    Args:
        query_id: The unique identifier of the persisted query.

    Returns:
        The query text.

    Raises:
        PersistedQueryError: If the query is not found or version mismatch.
    """
    _load_registry()

    if query_id not in _QUERY_CACHE:
        raise PersistedQueryError(
            message=f"Persisted query not found: {query_id}",
            query_id=query_id,
        )

    entry = _QUERY_CACHE[query_id]

    # Check schema version
    if entry.schema_version != SCHEMA_VERSION:
        raise PersistedQueryError(
            message=(
                f"Persisted query '{query_id}' was created for schema version "
                f"{entry.schema_version} but current version is {SCHEMA_VERSION}"
            ),
            query_id=query_id,
            schema_version=entry.schema_version,
        )

    return entry.query


def get_schema_version() -> str:
    """Get the current schema version."""
    return SCHEMA_VERSION


def list_persisted_queries() -> list[Dict[str, str]]:
    """
    List all available persisted queries.

    Returns:
        List of dicts with id, description, and schema_version.
    """
    _load_registry()

    return [
        {
            "id": entry.id,
            "description": entry.description,
            "schema_version": entry.schema_version,
        }
        for entry in _QUERY_CACHE.values()
    ]


def register_query(
    query_id: str,
    query: str,
    description: str = "",
) -> None:
    """
    Register a persisted query (in-memory only, for testing).

    Args:
        query_id: Unique identifier for the query.
        query: The GraphQL query text.
        description: Optional description.
    """
    _QUERY_CACHE[query_id] = PersistedQuery(
        id=query_id,
        query=query,
        schema_version=SCHEMA_VERSION,
        description=description,
    )


def clear_cache() -> None:
    """Clear the persisted query cache (for testing)."""
    global _QUERY_CACHE, _CACHE_LOADED
    _QUERY_CACHE = {}
    _CACHE_LOADED = False



"""Catalog resolver for GraphQL analytics API."""

from __future__ import annotations

import logging
from typing import Optional, TYPE_CHECKING

from ..authz import require_org_id
from ..context import GraphQLContext
from ..cost import DEFAULT_LIMITS
from ..loaders.dimension_loader import (
    get_dimension_descriptions,
    get_measure_descriptions,
    load_dimension_values,
)
from ..models.inputs import DimensionInput
from ..models.outputs import (
    CatalogDimension,
    CatalogLimits,
    CatalogMeasure,
    CatalogResult,
    CatalogValueItem,
)
from ..sql.validate import Dimension, Measure

if TYPE_CHECKING:
    from ..models.inputs import FilterInput


logger = logging.getLogger(__name__)


async def resolve_catalog(
    context: GraphQLContext,
    dimension: Optional[DimensionInput] = None,
    filters: Optional["FilterInput"] = None,  # NEW: Filter support
) -> CatalogResult:
    """
    Resolve catalog query.

    Returns available dimensions, measures, limits, and optionally
    distinct values for a specific dimension.

    Args:
        context: GraphQL request context with org_id.
        dimension: Optional dimension to fetch values for.
        filters: Optional filters to narrow down dimension values.

    Returns:
        CatalogResult with dimensions, measures, limits, and optional values.
    """
    org_id = require_org_id(context)

    # Build dimension list
    dim_descriptions = get_dimension_descriptions()
    dimensions = [
        CatalogDimension(name=d.value, description=dim_descriptions.get(d.value, ""))
        for d in Dimension
    ]

    # Build measure list
    measure_descriptions = get_measure_descriptions()
    measures = [
        CatalogMeasure(name=m.value, description=measure_descriptions.get(m.value, ""))
        for m in Measure
    ]

    # Build limits
    limits = CatalogLimits(
        max_days=DEFAULT_LIMITS.max_days,
        max_buckets=DEFAULT_LIMITS.max_buckets,
        max_top_n=DEFAULT_LIMITS.max_top_n,
        max_sankey_nodes=DEFAULT_LIMITS.max_sankey_nodes,
        max_sankey_edges=DEFAULT_LIMITS.max_sankey_edges,
        max_sub_requests=DEFAULT_LIMITS.max_sub_requests,
    )

    # Optionally fetch dimension values
    values = None
    if dimension is not None and context.client is not None:
        try:
            raw_values = await load_dimension_values(
                client=context.client,
                dimension=dimension.value,
                org_id=org_id,
                limit=100,
                filters=filters,
            )
            values = [
                CatalogValueItem(value=v["value"], count=v["count"]) for v in raw_values
            ]
        except Exception as e:
            logger.warning("Failed to load dimension values: %s", e)
            values = []

    return CatalogResult(
        dimensions=dimensions,
        measures=measures,
        limits=limits,
        values=values,
    )



from __future__ import annotations

from contextlib import asynccontextmanager
from datetime import date, datetime, timezone
import json

import pytest

from dev_health_ops.api.models.filters import MetricFilter, ScopeFilter, TimeFilter
from dev_health_ops.api.services.cache import TTLCache
from dev_health_ops.api.services import explain as explain_service
from dev_health_ops.api.services import home as home_service
from dev_health_ops.api.services import people as people_service


def _payload(model):
    if hasattr(model, "model_dump"):
        try:
            return model.model_dump(mode="json")
        except TypeError:
            return model.model_dump()
    return model.dict()


@asynccontextmanager
async def _fake_clickhouse_client(_dsn):
    yield object()


@pytest.mark.asyncio
async def test_home_response_sanitizes_non_finite_values(monkeypatch):
    async def _fake_scope_filter_for_metric(*_args, **_kwargs):
        return "", {}

    async def _fake_fetch_metric_value(*_args, **_kwargs):
        return float("nan")

    async def _fake_fetch_metric_series(*_args, **_kwargs):
        return [{"day": date(2024, 1, 1), "value": float("nan")}]

    async def _fake_fetch_blocked_hours(*_args, **_kwargs):
        return float("nan"), [{"day": date(2024, 1, 1), "value": float("nan")}]

    async def _fake_fetch_last_ingested_at(*_args, **_kwargs):
        return datetime(2024, 1, 1, tzinfo=timezone.utc)

    async def _fake_fetch_coverage(*_args, **_kwargs):
        return {
            "repos_covered_pct": 0.0,
            "prs_linked_to_issues_pct": 0.0,
            "issues_with_cycle_states_pct": 0.0,
        }

    async def _fake_fetch_metric_driver_delta(*_args, **_kwargs):
        return []

    monkeypatch.setattr(home_service, "clickhouse_client", _fake_clickhouse_client)
    monkeypatch.setattr(
        home_service, "scope_filter_for_metric", _fake_scope_filter_for_metric
    )
    monkeypatch.setattr(home_service, "fetch_metric_value", _fake_fetch_metric_value)
    monkeypatch.setattr(home_service, "fetch_metric_series", _fake_fetch_metric_series)
    monkeypatch.setattr(home_service, "fetch_blocked_hours", _fake_fetch_blocked_hours)
    monkeypatch.setattr(
        home_service, "fetch_last_ingested_at", _fake_fetch_last_ingested_at
    )
    monkeypatch.setattr(home_service, "fetch_coverage", _fake_fetch_coverage)
    monkeypatch.setattr(
        home_service, "fetch_metric_driver_delta", _fake_fetch_metric_driver_delta
    )

    response = await home_service.build_home_response(
        db_url="clickhouse://",
        filters=MetricFilter(
            time=TimeFilter(range_days=7, compare_days=7),
            scope=ScopeFilter(level="team", ids=["core"]),
        ),
        cache=TTLCache(ttl_seconds=1),
    )

    json.dumps(_payload(response), allow_nan=False)


@pytest.mark.asyncio
async def test_explain_response_sanitizes_non_finite_values(monkeypatch):
    async def _fake_scope_filter_for_metric(*_args, **_kwargs):
        return "", {}

    async def _fake_fetch_metric_value(*_args, **_kwargs):
        return float("nan")

    async def _fake_fetch_metric_driver_delta(*_args, **_kwargs):
        return [{"id": "team-a", "value": float("nan"), "delta_pct": float("nan")}]

    async def _fake_fetch_metric_contributors(*_args, **_kwargs):
        return [{"id": "team-b", "value": float("nan")}]

    monkeypatch.setattr(explain_service, "clickhouse_client", _fake_clickhouse_client)
    monkeypatch.setattr(
        explain_service, "scope_filter_for_metric", _fake_scope_filter_for_metric
    )
    monkeypatch.setattr(explain_service, "fetch_metric_value", _fake_fetch_metric_value)
    monkeypatch.setattr(
        explain_service, "fetch_metric_driver_delta", _fake_fetch_metric_driver_delta
    )
    monkeypatch.setattr(
        explain_service, "fetch_metric_contributors", _fake_fetch_metric_contributors
    )

    response = await explain_service.build_explain_response(
        db_url="clickhouse://",
        metric="wip_saturation",
        filters=MetricFilter(
            time=TimeFilter(range_days=7, compare_days=7),
            scope=ScopeFilter(level="team", ids=["core"]),
        ),
        cache=TTLCache(ttl_seconds=1),
    )

    json.dumps(_payload(response), allow_nan=False)


@pytest.mark.asyncio
async def test_person_summary_sanitizes_non_finite_values(monkeypatch):
    async def _fake_resolve_identity_context(*_args, **_kwargs):
        return "user@example.com", []

    async def _fake_fetch_last_ingested_at(*_args, **_kwargs):
        return datetime(2024, 1, 1, tzinfo=timezone.utc)

    async def _fake_fetch_coverage(*_args, **_kwargs):
        return {
            "repos_covered_pct": 0.0,
            "prs_linked_to_issues_pct": 0.0,
            "issues_with_cycle_states_pct": 0.0,
        }

    async def _fake_fetch_identity_coverage(*_args, **_kwargs):
        return float("nan")

    async def _fake_fetch_person_metric_value(*_args, **_kwargs):
        return float("nan")

    async def _fake_fetch_person_metric_series(*_args, **_kwargs):
        return [{"day": date(2024, 1, 1), "value": float("nan")}]

    async def _fake_fetch_person_work_mix(*_args, **_kwargs):
        return [{"key": "planned", "name": "Planned", "value": float("nan")}]

    async def _fake_fetch_person_flow_breakdown(*_args, **_kwargs):
        return [{"stage": "review", "value": float("nan"), "unit": "hours"}]

    async def _fake_fetch_person_collaboration(*_args, **_kwargs):
        return [{"section": "review_load", "label": "handoff", "value": float("nan")}]

    monkeypatch.setattr(people_service, "clickhouse_client", _fake_clickhouse_client)
    monkeypatch.setattr(
        people_service, "_resolve_identity_context", _fake_resolve_identity_context
    )
    monkeypatch.setattr(people_service, "load_identity_aliases", lambda: {})
    monkeypatch.setattr(
        people_service, "fetch_last_ingested_at", _fake_fetch_last_ingested_at
    )
    monkeypatch.setattr(people_service, "fetch_coverage", _fake_fetch_coverage)
    monkeypatch.setattr(
        people_service, "fetch_identity_coverage", _fake_fetch_identity_coverage
    )
    monkeypatch.setattr(
        people_service, "fetch_person_metric_value", _fake_fetch_person_metric_value
    )
    monkeypatch.setattr(
        people_service, "fetch_person_metric_series", _fake_fetch_person_metric_series
    )
    monkeypatch.setattr(
        people_service, "fetch_person_work_mix", _fake_fetch_person_work_mix
    )
    monkeypatch.setattr(
        people_service, "fetch_person_flow_breakdown", _fake_fetch_person_flow_breakdown
    )
    monkeypatch.setattr(
        people_service, "fetch_person_collaboration", _fake_fetch_person_collaboration
    )

    response = await people_service.build_person_summary_response(
        db_url="clickhouse://",
        person_id="person-1",
        range_days=7,
        compare_days=7,
    )

    json.dumps(_payload(response), allow_nan=False)


@pytest.mark.asyncio
async def test_person_metric_sanitizes_non_finite_values(monkeypatch):
    async def _fake_resolve_identity_context(*_args, **_kwargs):
        return "user@example.com", []

    async def _fake_fetch_person_metric_series(*_args, **_kwargs):
        return [{"day": date(2024, 1, 1), "value": float("nan")}]

    async def _fake_fetch_person_breakdown(*_args, **_kwargs):
        return [{"label": "Review", "value": float("nan")}]

    monkeypatch.setattr(people_service, "clickhouse_client", _fake_clickhouse_client)
    monkeypatch.setattr(
        people_service, "_resolve_identity_context", _fake_resolve_identity_context
    )
    monkeypatch.setattr(people_service, "load_identity_aliases", lambda: {})
    monkeypatch.setattr(
        people_service, "fetch_person_metric_series", _fake_fetch_person_metric_series
    )
    monkeypatch.setattr(
        people_service, "fetch_person_breakdown", _fake_fetch_person_breakdown
    )

    response = await people_service.build_person_metric_response(
        db_url="clickhouse://",
        person_id="person-1",
        metric="cycle_time",
        range_days=7,
        compare_days=7,
    )

    json.dumps(_payload(response), allow_nan=False)



from __future__ import annotations

from datetime import datetime, timezone

from dev_health_ops.providers.identity import IdentityResolver
from dev_health_ops.providers.jira.normalize import jira_issue_to_work_item
from dev_health_ops.providers.status_mapping import load_status_mapping


def test_jira_issue_dict_parses_status_category_and_changelog() -> None:
    status_mapping = load_status_mapping()
    identity = IdentityResolver(alias_to_canonical={})

    issue = {
        "key": "ABC-1",
        "self": "https://example.atlassian.net/rest/api/3/issue/10000",
        "fields": {
            "project": {"key": "ABC", "id": "10001"},
            "summary": "Test",
            "status": {"name": "Custom Done", "statusCategory": {"key": "done"}},
            "issuetype": {"name": "Task"},
            "labels": [],
            "created": "2025-12-01T10:00:00.000+0000",
            "updated": "2025-12-02T10:00:00.000+0000",
            "resolutiondate": "2025-12-02T10:00:00.000+0000",
        },
        "changelog": {
            "histories": [
                {
                    "created": "2025-12-01T12:00:00.000+0000",
                    "author": {"displayName": "Alice", "accountId": "a1"},
                    "items": [
                        {
                            "field": "status",
                            "fromString": "To Do",
                            "toString": "In Progress",
                        }
                    ],
                },
                {
                    "created": "2025-12-02T10:00:00.000+0000",
                    "author": {"displayName": "Alice", "accountId": "a1"},
                    "items": [
                        {
                            "field": "status",
                            "fromString": "In Progress",
                            "toString": "Custom Done",
                        }
                    ],
                },
            ]
        },
    }

    wi, transitions = jira_issue_to_work_item(
        issue=issue, status_mapping=status_mapping, identity=identity, repo_id=None
    )

    assert wi.work_item_id == "jira:ABC-1"
    assert wi.project_key == "ABC"
    # statusCategory=done should force normalized done even if status name is custom.
    assert wi.status == "done"
    assert wi.started_at == datetime(2025, 12, 1, 12, 0, tzinfo=timezone.utc)
    assert wi.completed_at == datetime(2025, 12, 2, 10, 0, tzinfo=timezone.utc)
    assert len(transitions) == 2



"""
Example script demonstrating access to private repositories on GitHub and GitLab.

This example shows how to properly configure tokens and access private repositories
using the GitHub and GitLab connectors.

Environment Variables:
    GITHUB_TOKEN: GitHub personal access token with 'repo' scope
    GITHUB_PRIVATE_REPO: Private repository in format 'owner/repo'
    GITLAB_TOKEN: GitLab private token with 'read_api' and 'read_repository' scopes
    GITLAB_PRIVATE_PROJECT: Private project name or ID
    GITLAB_URL: GitLab instance URL (default: https://gitlab.com)
"""

import os
import sys
import traceback

# Add parent directory to path for imports
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from dev_health_ops.connectors import GitHubConnector, GitLabConnector  # noqa: E402
from dev_health_ops.connectors.exceptions import (  # noqa: E402
    APIException,
    AuthenticationException,
)


def test_github_private_repo():
    """Test accessing a GitHub private repository."""
    print("\n" + "=" * 70)
    print("GitHub Private Repository Access Test")
    print("=" * 70)

    token = os.getenv("GITHUB_TOKEN")
    private_repo = os.getenv("GITHUB_PRIVATE_REPO")

    if not token:
        print("❌ GITHUB_TOKEN environment variable not set")
        print("   Please set it with: export GITHUB_TOKEN=your_token")
        print("   Token must have 'repo' scope for private repositories")
        return False

    if not private_repo:
        print("❌ GITHUB_PRIVATE_REPO environment variable not set")
        print("   Please set it with: export GITHUB_PRIVATE_REPO=owner/repo")
        return False

    try:
        owner, repo_name = private_repo.split("/")
    except ValueError:
        print(f"❌ Invalid format for GITHUB_PRIVATE_REPO: {private_repo}")
        print("   Should be in format 'owner/repo'")
        return False

    print(f"\nAttempting to access private repository: {private_repo}")
    print(f"Using token: {token[:10]}...")

    connector = GitHubConnector(token=token)

    try:
        # Test 1: List user's repositories (should include private ones)
        print("\n1. Listing user's repositories...")
        repos = connector.list_repositories(user_name=owner, max_repos=50)

        found = False
        for repo in repos:
            if repo.full_name == private_repo:
                found = True
                print(f"   ✅ Found private repository: {repo.full_name}")
                break

        if not found:
            print(f"   ⚠️  Private repository {private_repo} not found in user's repos")
            print("   This might mean the token doesn't have access to this repository")

        # Test 2: Get repository statistics
        print("\n2. Fetching repository statistics...")
        stats = connector.get_repo_stats(owner, repo_name, max_commits=10)
        print(f"   ✅ Total commits: {stats.total_commits}")
        print(f"   ✅ Total additions: {stats.additions}")
        print(f"   ✅ Total deletions: {stats.deletions}")
        print(f"   ✅ Authors: {len(stats.authors)}")

        # Test 3: Get contributors
        print("\n3. Fetching contributors...")
        contributors = connector.get_contributors(owner, repo_name, max_contributors=5)
        print(f"   ✅ Found {len(contributors)} contributors")
        for contributor in contributors[:3]:
            print(f"      - {contributor.username}")

        # Test 4: Check rate limit
        print("\n4. Checking rate limit...")
        rate_limit = connector.get_rate_limit()
        print(
            f"   ✅ Rate limit: {rate_limit['remaining']}/{rate_limit['limit']} remaining"
        )

        print("\n✅ Successfully accessed private GitHub repository!")
        return True

    except AuthenticationException as e:
        print(f"\n❌ Authentication failed: {e}")
        print("   Make sure your GITHUB_TOKEN has the 'repo' scope")
        print("   Go to https://github.com/settings/tokens to verify")
        return False

    except APIException as e:
        if "404" in str(e):
            print(f"\n❌ Repository not found: {e}")
            print(
                "   Either the repository doesn't exist or your token doesn't have access"
            )
        else:
            print(f"\n❌ API error: {e}")
        return False

    except Exception as e:
        print(f"\n❌ Unexpected error: {e}")
        traceback.print_exc()
        return False

    finally:
        connector.close()


def test_gitlab_private_project():
    """Test accessing a GitLab private project."""
    print("\n" + "=" * 70)
    print("GitLab Private Project Access Test")
    print("=" * 70)

    token = os.getenv("GITLAB_TOKEN")
    private_project = os.getenv("GITLAB_PRIVATE_PROJECT")
    gitlab_url = os.getenv("GITLAB_URL", "https://gitlab.com")

    if not token:
        print("❌ GITLAB_TOKEN environment variable not set")
        print("   Please set it with: export GITLAB_TOKEN=your_token")
        print("   Token must have 'read_api' and 'read_repository' scopes")
        return False

    if not private_project:
        print("❌ GITLAB_PRIVATE_PROJECT environment variable not set")
        print("   Please set it with: export GITLAB_PRIVATE_PROJECT=group/project")
        print("   Or use a project ID: export GITLAB_PRIVATE_PROJECT=12345")
        return False

    print(f"\nAttempting to access private project: {private_project}")
    print(f"GitLab URL: {gitlab_url}")
    print(f"Using token: {token[:10]}...")

    connector = GitLabConnector(url=gitlab_url, private_token=token)

    try:
        # Test 1: Get project details
        print("\n1. Fetching project details...")
        try:
            # Direct python-gitlab API call to verify basic access
            project = connector.gitlab.projects.get(private_project)
            print(f"   ✅ Found project: {project.name}")
            print(
                f"   ✅ Full path: {project.path_with_namespace if hasattr(project, 'path_with_namespace') else 'N/A'}"
            )
        except Exception as e:
            # python-gitlab can raise various exceptions depending on the error
            print(f"   ❌ Failed to access project: {e}")
            return False

        # Determine project identifier for subsequent calls
        project_identifier = private_project

        # Test 2: Get project statistics
        print("\n2. Fetching project statistics...")
        try:
            stats = connector.get_repo_stats(
                project_name=project_identifier, max_commits=10
            )
        except Exception:
            if str(private_project).isdigit():
                stats = connector.get_repo_stats(
                    project_id=int(private_project), max_commits=10
                )
            else:
                raise

        print(f"   ✅ Total commits: {stats.total_commits}")
        print(f"   ✅ Total additions: {stats.additions}")
        print(f"   ✅ Total deletions: {stats.deletions}")
        print(f"   ✅ Authors: {len(stats.authors)}")

        # Test 3: Get contributors
        print("\n3. Fetching contributors...")
        try:
            contributors = connector.get_contributors(
                project_name=project_identifier, max_contributors=5
            )
        except Exception:
            if str(private_project).isdigit():
                contributors = connector.get_contributors(
                    project_id=int(private_project), max_contributors=5
                )
            else:
                raise

        print(f"   ✅ Found {len(contributors)} contributors")
        for contributor in contributors[:3]:
            print(f"      - {contributor.username}")

        # Test 4: List user's projects (should include private ones)
        print("\n4. Listing accessible projects...")
        projects = connector.list_projects(max_projects=10)
        print(f"   ✅ Found {len(projects)} accessible projects (may include private)")

        print("\n✅ Successfully accessed private GitLab project!")
        return True

    except AuthenticationException as e:
        print(f"\n❌ Authentication failed: {e}")
        print(
            "   Make sure your GITLAB_TOKEN has 'read_api' and 'read_repository' scopes"
        )
        return False

    except APIException as e:
        if "404" in str(e):
            print(f"\n❌ Project not found: {e}")
            print(
                "   Either the project doesn't exist or your token doesn't have access"
            )
        else:
            print(f"\n❌ API error: {e}")
        return False

    except Exception as e:
        print(f"\n❌ Unexpected error: {e}")
        traceback.print_exc()
        return False

    finally:
        connector.close()


def main():
    """Main function to run all tests."""
    print("\n" + "=" * 70)
    print("Private Repository Access Test Suite")
    print("=" * 70)
    print("\nThis script tests access to private repositories on GitHub and GitLab.")
    print("Make sure you have set the required environment variables:")
    print("\nFor GitHub:")
    print("  export GITHUB_TOKEN=your_token")
    print("  export GITHUB_PRIVATE_REPO=owner/repo")
    print("\nFor GitLab:")
    print("  export GITLAB_TOKEN=your_token")
    print("  export GITLAB_PRIVATE_PROJECT=group/project  # or project ID")
    print("  export GITLAB_URL=https://gitlab.com  # optional")

    results = []

    # Test GitHub
    if os.getenv("GITHUB_TOKEN") and os.getenv("GITHUB_PRIVATE_REPO"):
        results.append(("GitHub", test_github_private_repo()))
    else:
        print("\n⏭️  Skipping GitHub test (missing GITHUB_TOKEN or GITHUB_PRIVATE_REPO)")
        results.append(("GitHub", None))

    # Test GitLab
    if os.getenv("GITLAB_TOKEN") and os.getenv("GITLAB_PRIVATE_PROJECT"):
        results.append(("GitLab", test_gitlab_private_project()))
    else:
        print(
            "\n⏭️  Skipping GitLab test (missing GITLAB_TOKEN or GITLAB_PRIVATE_PROJECT)"
        )
        results.append(("GitLab", None))

    # Summary
    print("\n" + "=" * 70)
    print("Test Summary")
    print("=" * 70)

    for service, result in results:
        if result is None:
            status = "⏭️  SKIPPED"
        elif result:
            status = "✅ PASSED"
        else:
            status = "❌ FAILED"
        print(f"{service}: {status}")

    # Exit with appropriate code
    failed = any(result is False for _, result in results)
    if failed:
        print("\n❌ Some tests failed. Check the output above for details.")
        sys.exit(1)

    ran_tests = any(result is not None for _, result in results)
    if not ran_tests:
        print("\n⚠️  No tests were run. Set environment variables to run tests.")
        sys.exit(0)

    print("\n✅ All tests passed!")
    sys.exit(0)


if __name__ == "__main__":
    main()



"""
GitHub and GitLab connectors for retrieving repository data.

This package provides production-grade connectors for GitHub and GitLab
with automatic pagination, rate limiting, and error handling.
"""

from .base import BatchResult, GitConnector
from .exceptions import (
    APIException,
    AuthenticationException,
    ConnectorException,
    NotFoundException,
    PaginationException,
    RateLimitException,
)
from .github import GitHubConnector, match_repo_pattern
from .gitlab import GitLabConnector, match_project_pattern
from .models import (
    Author,
    BlameRange,
    CommitStats,
    FileBlame,
    Organization,
    PullRequest,
    Repository,
    RepoStats,
)

__all__ = [
    # Base class
    "GitConnector",
    # Connectors
    "GitHubConnector",
    "GitLabConnector",
    # Batch processing
    "BatchResult",
    "match_repo_pattern",
    "match_project_pattern",
    # Models
    "Organization",
    "Repository",
    "Author",
    "CommitStats",
    "RepoStats",
    "PullRequest",
    "BlameRange",
    "FileBlame",
    # Exceptions
    "ConnectorException",
    "RateLimitException",
    "AuthenticationException",
    "NotFoundException",
    "PaginationException",
    "APIException",
]



from __future__ import annotations

from datetime import datetime
from typing import Any, Dict, Iterable, List, Optional, Tuple

from .client import query_dicts


async def fetch_work_unit_investments(
    client: Any,
    *,
    start_ts: datetime,
    end_ts: datetime,
    repo_ids: Optional[List[str]],
    limit: int,
    work_unit_id: Optional[str] = None,
) -> List[Dict[str, Any]]:
    params: Dict[str, Any] = {"start_ts": start_ts, "end_ts": end_ts, "limit": limit}
    # ClickHouse may prefer alias over column names in WHERE; always qualify columns
    # to avoid accidentally referencing argMax(...) aliases.
    filters: List[str] = [
        "work_unit_investments.from_ts < %(end_ts)s",
        "work_unit_investments.to_ts >= %(start_ts)s",
    ]
    if repo_ids:
        filters.append("work_unit_investments.repo_id IN %(repo_ids)s")
        params["repo_ids"] = repo_ids
    if work_unit_id:
        filters.append("work_unit_investments.work_unit_id = %(work_unit_id)s")
        params["work_unit_id"] = work_unit_id
    where_sql = " AND ".join(filters)
    query = f"""
        SELECT
            work_unit_id,
            argMax(work_unit_type, work_unit_investments.computed_at) AS work_unit_type,
            argMax(work_unit_name, work_unit_investments.computed_at) AS work_unit_name,
            argMax(from_ts, work_unit_investments.computed_at) AS from_ts,
            argMax(to_ts, work_unit_investments.computed_at) AS to_ts,
            argMax(repo_id, work_unit_investments.computed_at) AS repo_id,
            argMax(provider, work_unit_investments.computed_at) AS provider,
            argMax(effort_metric, work_unit_investments.computed_at) AS effort_metric,
            argMax(effort_value, work_unit_investments.computed_at) AS effort_value,
            argMax(theme_distribution_json, work_unit_investments.computed_at) AS theme_distribution_json,
            argMax(subcategory_distribution_json, work_unit_investments.computed_at) AS subcategory_distribution_json,
            argMax(structural_evidence_json, work_unit_investments.computed_at) AS structural_evidence_json,
            argMax(evidence_quality, work_unit_investments.computed_at) AS evidence_quality,
            argMax(evidence_quality_band, work_unit_investments.computed_at) AS evidence_quality_band,
            argMax(categorization_status, work_unit_investments.computed_at) AS categorization_status,
            argMax(categorization_run_id, work_unit_investments.computed_at) AS categorization_run_id,
            max(work_unit_investments.computed_at) AS computed_at
        FROM work_unit_investments
        WHERE {where_sql}
        GROUP BY work_unit_id
        ORDER BY effort_value DESC
        LIMIT %(limit)s
    """
    return await query_dicts(client, query, params)


async def fetch_repo_scopes(
    client: Any,
    *,
    repo_ids: Iterable[str],
) -> Dict[str, str]:
    ids = [repo_id for repo_id in repo_ids if repo_id]
    if not ids:
        return {}
    query = """
        SELECT
            toString(id) AS repo_id,
            repo
        FROM repos
        WHERE id IN %(repo_ids)s
    """
    rows = await query_dicts(client, query, {"repo_ids": ids})
    return {
        str(row.get("repo_id")): str(row.get("repo") or "")
        for row in rows
        if row.get("repo_id")
    }


async def fetch_work_item_team_assignments(
    client: Any,
    *,
    work_item_ids: Iterable[str],
) -> Dict[str, Dict[str, str]]:
    ids = [work_item_id for work_item_id in work_item_ids if work_item_id]
    if not ids:
        return {}
    query = """
        SELECT
            work_item_id,
            argMax(team_id, computed_at) AS team_id,
            argMax(team_name, computed_at) AS team_name
        FROM work_item_cycle_times
        WHERE work_item_id IN %(work_item_ids)s
        GROUP BY work_item_id
    """
    rows = await query_dicts(client, query, {"work_item_ids": ids})
    result: Dict[str, Dict[str, str]] = {}
    for row in rows:
        work_item_id = str(row.get("work_item_id") or "")
        if not work_item_id:
            continue
        team_id = str(row.get("team_id") or "")
        team_name = str(row.get("team_name") or "")
        result[work_item_id] = {"team_id": team_id, "team_name": team_name}
    return result


async def fetch_work_unit_investment_quotes(
    client: Any,
    *,
    unit_runs: Iterable[Tuple[str, str]],
) -> List[Dict[str, Any]]:
    pairs = [(unit_id, run_id) for unit_id, run_id in unit_runs if unit_id and run_id]
    if not pairs:
        return []
    query = """
        SELECT
            work_unit_id,
            quote,
            source_type,
            source_id,
            categorization_run_id
        FROM work_unit_investment_quotes
        WHERE (work_unit_id, categorization_run_id) IN %(pairs)s
    """
    return await query_dicts(client, query, {"pairs": pairs})



import argparse
import asyncio
import os

from dev_health_ops.storage import resolve_db_type, run_with_store
from dev_health_ops.utils import _resolve_since, _resolve_max_commits
from dev_health_ops.processors.local import process_local_blame, process_local_repo
from dev_health_ops.processors.github import (
    process_github_repo,
    process_github_repos_batch,
)
from dev_health_ops.processors.gitlab import (
    process_gitlab_project,
    process_gitlab_projects_batch,
)


def _sync_flags_for_target(target: str) -> dict:
    return {
        "sync_git": target == "git",
        "sync_prs": target == "prs",
        "sync_cicd": target == "cicd",
        "sync_deployments": target == "deployments",
        "sync_incidents": target == "incidents",
        "blame_only": target == "blame",
    }


def _resolve_synthetic_repo_name(ns: argparse.Namespace) -> str:
    if ns.repo_name:
        return ns.repo_name
    if ns.owner and ns.repo:
        return f"{ns.owner}/{ns.repo}"
    if ns.search:
        if "*" in ns.search or "?" in ns.search:
            raise SystemExit(
                "Synthetic provider does not support pattern search; use --repo-name."
            )
        return ns.search
    return "acme/demo-app"


async def sync_local_target(ns: argparse.Namespace, target: str) -> int:
    if target not in {"git", "prs", "blame"}:
        raise SystemExit("Local provider supports only git, prs, or blame targets.")

    db_type = resolve_db_type(ns.db, ns.db_type)
    since = _resolve_since(ns)

    async def _handler(store):
        if target == "blame":
            await process_local_blame(
                store=store,
                repo_path=ns.repo_path,
                since=since,
            )
            return

        await process_local_repo(
            store=store,
            repo_path=ns.repo_path,
            since=since,
            sync_git=(target == "git"),
            sync_prs=(target == "prs"),
            sync_blame=False,
        )

    await run_with_store(ns.db, db_type, _handler)
    return 0


async def sync_github_target(ns: argparse.Namespace, target: str) -> int:
    token = ns.auth or os.getenv("GITHUB_TOKEN") or ""
    if not token:
        raise SystemExit("Missing GitHub token (pass --auth or set GITHUB_TOKEN).")

    db_type = resolve_db_type(ns.db, ns.db_type)
    since = _resolve_since(ns)
    max_commits = _resolve_max_commits(ns)
    flags = _sync_flags_for_target(target)

    async def _handler(store):
        if ns.search:
            org_name = ns.group
            user_name = ns.owner if not ns.group else None
            await process_github_repos_batch(
                store=store,
                token=token,
                org_name=org_name,
                user_name=user_name,
                pattern=ns.search,
                batch_size=ns.batch_size,
                max_concurrent=ns.max_concurrent,
                rate_limit_delay=ns.rate_limit_delay,
                max_commits_per_repo=max_commits,
                max_repos=ns.max_repos,
                use_async=ns.use_async,
                sync_git=flags["sync_git"],
                sync_prs=flags["sync_prs"],
                sync_cicd=flags["sync_cicd"],
                sync_deployments=flags["sync_deployments"],
                sync_incidents=flags["sync_incidents"],
                blame_only=flags["blame_only"],
                backfill_missing=False,
                since=since,
            )
            return

        if not (ns.owner and ns.repo):
            raise SystemExit(
                "GitHub sync requires --owner and --repo (or --search for batch)."
            )
        await process_github_repo(
            store,
            ns.owner,
            ns.repo,
            token,
            blame_only=flags["blame_only"],
            max_commits=max_commits,
            sync_git=flags["sync_git"],
            sync_prs=flags["sync_prs"],
            sync_cicd=flags["sync_cicd"],
            sync_deployments=flags["sync_deployments"],
            sync_incidents=flags["sync_incidents"],
            since=since,
        )

    await run_with_store(ns.db, db_type, _handler)
    return 0


async def sync_gitlab_target(ns: argparse.Namespace, target: str) -> int:
    token = ns.auth or os.getenv("GITLAB_TOKEN") or ""
    if not token:
        raise SystemExit("Missing GitLab token (pass --auth or set GITLAB_TOKEN).")

    db_type = resolve_db_type(ns.db, ns.db_type)
    since = _resolve_since(ns)
    max_commits = _resolve_max_commits(ns)
    flags = _sync_flags_for_target(target)

    async def _handler(store):
        if ns.search:
            await process_gitlab_projects_batch(
                store=store,
                token=token,
                gitlab_url=ns.gitlab_url,
                group_name=ns.group,
                pattern=ns.search,
                batch_size=ns.batch_size,
                max_concurrent=ns.max_concurrent,
                rate_limit_delay=ns.rate_limit_delay,
                max_commits_per_project=max_commits,
                max_projects=ns.max_repos,
                use_async=ns.use_async,
                sync_git=flags["sync_git"],
                sync_prs=flags["sync_prs"],
                sync_cicd=flags["sync_cicd"],
                sync_deployments=flags["sync_deployments"],
                sync_incidents=flags["sync_incidents"],
                blame_only=flags["blame_only"],
                backfill_missing=False,
                since=since,
            )
            return

        if ns.project_id is None:
            raise SystemExit(
                "GitLab sync requires --project-id (or --search for batch)."
            )
        await process_gitlab_project(
            store,
            ns.project_id,
            token,
            ns.gitlab_url,
            blame_only=flags["blame_only"],
            max_commits=max_commits,
            sync_git=flags["sync_git"],
            sync_prs=flags["sync_prs"],
            sync_cicd=flags["sync_cicd"],
            sync_deployments=flags["sync_deployments"],
            sync_incidents=flags["sync_incidents"],
            since=since,
        )

    await run_with_store(ns.db, db_type, _handler)
    return 0


async def sync_synthetic_target(ns: argparse.Namespace, target: str) -> int:
    from dev_health_ops.fixtures.generator import SyntheticDataGenerator

    repo_name = _resolve_synthetic_repo_name(ns)
    db_type = resolve_db_type(ns.db, ns.db_type)
    days = max(1, int(ns.backfill))

    async def _handler(store):
        generator = SyntheticDataGenerator(repo_name=repo_name)
        repo = generator.generate_repo()
        await store.insert_repo(repo)

        if target == "git":
            commits = generator.generate_commits(days=days)
            await store.insert_git_commit_data(commits)
            stats = generator.generate_commit_stats(commits)
            await store.insert_git_commit_stats(stats)
            return

        if target == "prs":
            pr_data = generator.generate_prs()
            prs = [p["pr"] for p in pr_data]
            await store.insert_git_pull_requests(prs)

            reviews = []
            for p in pr_data:
                reviews.extend(p["reviews"])
            if reviews:
                await store.insert_git_pull_request_reviews(reviews)
            return

        if target == "blame":
            commits = generator.generate_commits(days=days)
            files = generator.generate_files()
            await store.insert_git_file_data(files)
            blame_data = generator.generate_blame(commits)
            if blame_data:
                await store.insert_blame_data(blame_data)
            return

    await run_with_store(ns.db, db_type, _handler)
    return 0


def run_sync_target(ns: argparse.Namespace) -> int:
    target = ns.sync_target
    provider = (ns.provider or "").lower()
    if provider not in {"local", "github", "gitlab", "synthetic"}:
        raise SystemExit("Provider must be one of: local, github, gitlab, synthetic.")

    if target not in {"git", "prs", "blame", "cicd", "deployments", "incidents"}:
        raise SystemExit(
            "Sync target must be git, prs, blame, cicd, deployments, or incidents."
        )

    if provider == "local":
        return asyncio.run(sync_local_target(ns, target))
    if provider == "github":
        return asyncio.run(sync_github_target(ns, target))
    if provider == "gitlab":
        return asyncio.run(sync_gitlab_target(ns, target))
    return asyncio.run(sync_synthetic_target(ns, target))


def _add_sync_target_args(parser: argparse.ArgumentParser) -> None:
    parser.add_argument("--db", required=True, help="Database connection string.")
    parser.add_argument(
        "--db-type",
        choices=["postgres", "mongo", "sqlite", "clickhouse"],
        help="Optional DB backend override.",
    )
    parser.add_argument(
        "--provider",
        choices=["local", "github", "gitlab", "synthetic"],
        required=True,
        help="Source provider for the sync job.",
    )
    parser.add_argument("--auth", help="Provider token override (GitHub/GitLab).")
    parser.add_argument(
        "--repo-path", default=".", help="Local git repo path (local provider)."
    )
    parser.add_argument("--owner", help="GitHub owner/org (single repo mode).")
    parser.add_argument("--repo", help="GitHub repo name (single repo mode).")
    parser.add_argument(
        "--project-id", type=int, help="GitLab project ID (single project mode)."
    )
    parser.add_argument(
        "--gitlab-url",
        default=os.getenv("GITLAB_URL", "https://gitlab.com"),
        help="GitLab instance URL.",
    )
    parser.add_argument("--group", help="Batch mode org/group name.")
    parser.add_argument(
        "-s",
        "--search",
        help="Batch mode pattern (e.g. 'org/*').",
    )
    parser.add_argument("--batch-size", type=int, default=10)
    parser.add_argument("--max-concurrent", type=int, default=4)
    parser.add_argument("--rate-limit-delay", type=float, default=1.0)
    parser.add_argument("--max-repos", type=int)
    parser.add_argument("--use-async", action="store_true")
    parser.add_argument("--max-commits-per-repo", type=int)
    parser.add_argument(
        "--repo-name", help="Synthetic repo name (default: acme/demo-app)."
    )
    time_group = parser.add_mutually_exclusive_group()
    time_group.add_argument("--since", help="Lower-bound ISO date/time (UTC).")
    time_group.add_argument(
        "--date",
        help="Target day (UTC) as YYYY-MM-DD (use with --backfill).",
    )
    parser.add_argument(
        "--backfill",
        type=int,
        default=1,
        help="Sync N days ending at --date (inclusive). Requires --date.",
    )


def register_commands(subparsers: argparse._SubParsersAction) -> None:
    target_parsers = {
        "git": "Sync commits and commit stats.",
        "prs": "Sync pull/merge requests.",
        "blame": "Sync blame data only.",
        "cicd": "Sync CI/CD runs and pipelines.",
        "deployments": "Sync deployments.",
        "incidents": "Sync incidents.",
    }

    for target, help_text in target_parsers.items():
        target_parser = subparsers.add_parser(target, help=help_text)
        _add_sync_target_args(target_parser)
        target_parser.set_defaults(func=run_sync_target, sync_target=target)

    # Note: 'teams' and 'work-items' are also sync subcommands but handled in their own modules.



"""ClickHouse data loader implementation."""

from __future__ import annotations

import uuid
from datetime import datetime
from typing import Any, Dict, List, Optional, Tuple

from dev_health_ops.metrics.loaders.base import (
    DataLoader,
    naive_utc,
    parse_uuid,
    to_dataclass,
)
from dev_health_ops.metrics.schemas import (
    CommitStatRow,
    PullRequestRow,
    PullRequestReviewRow,
    PipelineRunRow,
    DeploymentRow,
    IncidentRow,
)


async def _clickhouse_query_dicts(
    client: Any, query: str, params: Dict[str, Any]
) -> List[Dict[str, Any]]:
    from dev_health_ops.api.queries.client import query_dicts

    return await query_dicts(client, query, params)


class ClickHouseDataLoader(DataLoader):
    """DataLoader implementation for ClickHouse backend."""

    def __init__(self, client: Any) -> None:
        self.client = client

    async def load_git_rows(
        self,
        start: datetime,
        end: datetime,
        repo_id: Optional[uuid.UUID],
        repo_name: Optional[str] = None,
    ) -> Tuple[List[CommitStatRow], List[PullRequestRow], List[PullRequestReviewRow]]:
        params: Dict[str, Any] = {"start": naive_utc(start), "end": naive_utc(end)}
        repo_filter = ""
        if repo_id is not None:
            params["repo_id"] = str(repo_id)
            repo_filter = " AND c.repo_id = {repo_id:UUID}"
        elif repo_name is not None:
            params["repo_name"] = repo_name
            repo_filter = " AND c.repo_id IN (SELECT id FROM repos WHERE repo = {repo_name:String})"

        commit_query = f"""
        SELECT
          c.repo_id AS repo_id,
          c.hash AS commit_hash,
          c.author_email AS author_email,
          c.author_name AS author_name,
          c.committer_when AS committer_when,
          s.file_path AS file_path,
          s.additions AS additions,
          s.deletions AS deletions
        FROM git_commits AS c
        LEFT JOIN git_commit_stats AS s
          ON (s.repo_id = c.repo_id) AND (s.commit_hash = c.hash)
        WHERE c.committer_when >= {{start:DateTime}} AND c.committer_when < {{end:DateTime}}
        {repo_filter}
        """

        pr_query = f"""
        SELECT
          repo_id,
          number,
          author_email,
          author_name,
          created_at,
          merged_at,
          first_review_at,
          first_comment_at,
          changes_requested_count,
          reviews_count,
          comments_count,
          additions,
          deletions,
          changed_files
        FROM git_pull_requests
        WHERE
          (created_at >= {{start:DateTime}} AND created_at < {{end:DateTime}})
          OR (merged_at IS NOT NULL AND merged_at >= {{start:DateTime}} AND merged_at < {{end:DateTime}})
          {repo_filter.replace("c.repo_id", "repo_id") if repo_id or repo_name else ""}
        """

        review_query = f"""
        SELECT
          repo_id,
          number,
          reviewer,
          submitted_at,
          state
        FROM git_pull_request_reviews
        WHERE submitted_at >= {{start:DateTime}} AND submitted_at < {{end:DateTime}}
        {repo_filter.replace("c.repo_id", "repo_id") if repo_id or repo_name else ""}
        """

        commit_dicts = await _clickhouse_query_dicts(self.client, commit_query, params)
        pr_dicts = await _clickhouse_query_dicts(self.client, pr_query, params)
        review_dicts = await _clickhouse_query_dicts(self.client, review_query, params)

        commit_rows: List[CommitStatRow] = []
        for r in commit_dicts:
            u = parse_uuid(r.get("repo_id"))
            cw = r.get("committer_when")
            if u and cw:
                commit_rows.append(
                    {
                        "repo_id": u,
                        "commit_hash": str(r.get("commit_hash") or ""),
                        "author_email": r.get("author_email") or "",
                        "author_name": r.get("author_name") or "",
                        "committer_when": cw,
                        "file_path": r.get("file_path"),
                        "additions": int(r.get("additions") or 0),
                        "deletions": int(r.get("deletions") or 0),
                    }
                )

        pr_rows: List[PullRequestRow] = []
        for r in pr_dicts:
            u = parse_uuid(r.get("repo_id"))
            ca = r.get("created_at")
            if u and ca:
                pr_rows.append(
                    {
                        "repo_id": u,
                        "number": int(r.get("number") or 0),
                        "author_email": r.get("author_email") or "",
                        "author_name": r.get("author_name") or "",
                        "created_at": ca,
                        "merged_at": r.get("merged_at"),
                        "first_review_at": r.get("first_review_at"),
                        "first_comment_at": r.get("first_comment_at"),
                        "changes_requested_count": int(
                            r.get("changes_requested_count", 0)
                        ),
                        "reviews_count": int(r.get("reviews_count", 0)),
                        "comments_count": int(r.get("comments_count", 0)),
                        "additions": int(r.get("additions", 0)),
                        "deletions": int(r.get("deletions", 0)),
                        "changed_files": int(r.get("changed_files", 0)),
                    }
                )

        review_rows: List[PullRequestReviewRow] = []
        for r in review_dicts:
            u = parse_uuid(r.get("repo_id"))
            sa = r.get("submitted_at")
            if u and sa:
                review_rows.append(
                    {
                        "repo_id": u,
                        "number": int(r.get("number") or 0),
                        "reviewer": r.get("reviewer") or "unknown",
                        "submitted_at": sa,
                        "state": r.get("state") or "unknown",
                    }
                )

        return commit_rows, pr_rows, review_rows

    async def load_work_items(
        self,
        start: datetime,
        end: datetime,
        repo_id: Optional[uuid.UUID],
        repo_name: Optional[str] = None,
    ) -> Tuple[List[Any], List[Any]]:
        from dev_health_ops.models.work_items import WorkItem, WorkItemStatusTransition

        params: Dict[str, Any] = {"start": naive_utc(start), "end": naive_utc(end)}
        repo_filter = ""
        if repo_id is not None:
            params["repo_id"] = str(repo_id)
            repo_filter = " AND repo_id = {repo_id:UUID}"

        item_query = f"""
        SELECT * FROM work_items
        WHERE (created_at < {{end:DateTime}})
        AND (status != 'done' OR completed_at >= {{start:DateTime}})
        {repo_filter}
        """

        trans_query = f"""
        SELECT * FROM work_item_transitions
        WHERE (occurred_at < {{end:DateTime}})
        {repo_filter}
        """

        item_dicts = await _clickhouse_query_dicts(self.client, item_query, params)
        trans_dicts = await _clickhouse_query_dicts(self.client, trans_query, params)

        items = [to_dataclass(WorkItem, d) for d in item_dicts]
        transitions = [to_dataclass(WorkItemStatusTransition, t) for t in trans_dicts]

        return items, transitions

    async def load_cicd_data(
        self,
        start: datetime,
        end: datetime,
        repo_id: Optional[uuid.UUID],
        repo_name: Optional[str] = None,
    ) -> Tuple[List[PipelineRunRow], List[DeploymentRow]]:
        params: Dict[str, Any] = {"start": naive_utc(start), "end": naive_utc(end)}
        repo_filter = ""
        if repo_id is not None:
            params["repo_id"] = str(repo_id)
            repo_filter = " AND repo_id = {repo_id:UUID}"

        pipe_query = f"""
        SELECT * FROM ci_pipeline_runs
        WHERE finished_at >= {{start:DateTime}} AND finished_at < {{end:DateTime}}
        {repo_filter}
        """
        deploy_query = f"""
        SELECT * FROM deployments
        WHERE deployed_at >= {{start:DateTime}} AND deployed_at < {{end:DateTime}}
        {repo_filter}
        """

        pipes_dicts = await _clickhouse_query_dicts(self.client, pipe_query, params)
        deploys_dicts = await _clickhouse_query_dicts(self.client, deploy_query, params)

        # ClickHouse dicts can be directly cast if they match keys
        pipes: List[PipelineRunRow] = [dict(p) for p in pipes_dicts]  # type: ignore
        deploys: List[DeploymentRow] = [dict(d) for d in deploys_dicts]  # type: ignore

        return pipes, deploys

    async def load_incidents(
        self,
        start: datetime,
        end: datetime,
        repo_id: Optional[uuid.UUID],
        repo_name: Optional[str] = None,
    ) -> List[IncidentRow]:
        params: Dict[str, Any] = {"start": naive_utc(start), "end": naive_utc(end)}
        repo_filter = ""
        if repo_id is not None:
            params["repo_id"] = str(repo_id)
            repo_filter = " AND repo_id = {repo_id:UUID}"

        query = f"""
        SELECT * FROM incidents
        WHERE started_at >= {{start:DateTime}} AND started_at < {{end:DateTime}}
        {repo_filter}
        """
        dicts = await _clickhouse_query_dicts(self.client, query, params)
        return [dict(d) for d in dicts]  # type: ignore

    async def load_blame_concentration(
        self,
        repo_id: uuid.UUID,
        as_of: datetime,
    ) -> Dict[uuid.UUID, float]:
        params = {"repo_id": str(repo_id), "as_of": naive_utc(as_of)}
        query = """
        SELECT
            repo_id,
            sum(lines_count * lines_count) / (sum(lines_count) * sum(lines_count)) as concentration
        FROM git_file_blame
        WHERE repo_id = {repo_id:UUID}
        GROUP BY repo_id
        """
        rows = await _clickhouse_query_dicts(self.client, query, params)
        res = {}
        for r in rows:
            u = parse_uuid(r.get("repo_id"))
            if u:
                res[u] = float(r["concentration"])
        return res



from __future__ import annotations

from dataclasses import dataclass
from datetime import date, datetime, time, timezone
import logging
from typing import Any, Dict, List, Optional, Tuple

from ..models.filters import MetricFilter, SankeyContext
from ..models.schemas import SankeyLink, SankeyNode, SankeyResponse
from ..queries.client import clickhouse_client, query_dicts
from ..queries.sankey import (
    fetch_expense_abandoned,
    fetch_expense_counts,
    fetch_hotspot_rows,
    fetch_investment_flow_items,
    fetch_state_status_counts,
)
from ..queries.scopes import build_scope_filter_multi
from .filtering import resolve_repo_filter_ids, time_window

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class SankeyDefinition:
    label: str
    description: str
    unit: str


SANKEY_DEFINITIONS: Dict[str, SankeyDefinition] = {
    "investment": SankeyDefinition(
        label="Investment flow",
        description=(
            "Where effort allocates across initiatives, areas, issue types, and work items."
        ),
        unit="items",
    ),
    "expense": SankeyDefinition(
        label="Investment expense",
        description="How planned effort converts into unplanned work, rework, and rewrites.",
        unit="items",
    ),
    "state": SankeyDefinition(
        label="State flow",
        description="Execution paths that reveal stalls, loops, and retry patterns.",
        unit="items",
    ),
    "hotspot": SankeyDefinition(
        label="Code hotspot flow",
        description="Where change concentrates from repos to files and change intent.",
        unit="changes",
    ),
}

MAX_INVESTMENT_ITEMS = 60
MAX_HOTSPOT_ROWS = 150


def _apply_window_to_filters(
    filters: MetricFilter,
    window_start: Optional[date],
    window_end: Optional[date],
) -> MetricFilter:
    if not window_start and not window_end:
        return filters
    payload = (
        filters.model_dump(mode="json")
        if hasattr(filters, "model_dump")
        else filters.dict()
    )
    time_payload = payload.get("time") or {}
    if window_start:
        time_payload["start_date"] = window_start
    if window_end:
        time_payload["end_date"] = window_end
    if window_start and window_end:
        delta_days = max(1, (window_end - window_start).days)
        time_payload["range_days"] = delta_days
    payload["time"] = time_payload
    return MetricFilter(**payload)


def _normalize_label(value: Any, fallback: str) -> str:
    if value is None:
        return fallback
    text = str(value).strip()
    return text or fallback


def _add_edge(
    edges: Dict[Tuple[str, str], float],
    source: str,
    target: str,
    value: float,
) -> None:
    if value <= 0:
        return
    key = (source, target)
    edges[key] = edges.get(key, 0.0) + value


def _touch_node(
    nodes: Dict[str, SankeyNode],
    name: str,
    group: Optional[str],
) -> None:
    if name in nodes:
        return
    nodes[name] = SankeyNode(name=name, group=group)


def _links_from_edges(edges: Dict[Tuple[str, str], float]) -> List[SankeyLink]:
    links = [
        SankeyLink(source=source, target=target, value=value)
        for (source, target), value in edges.items()
        if value > 0
    ]
    links.sort(key=lambda link: link.value, reverse=True)
    return links


async def _tables_present(client: Any, tables: List[str]) -> bool:
    if not tables:
        return True
    try:
        rows = await query_dicts(
            client,
            """
            SELECT name
            FROM system.tables
            WHERE database = currentDatabase()
                AND name IN %(tables)s
            """,
            {"tables": tables},
        )
    except Exception as exc:
        logger.warning("Sankey table lookup failed: %s", exc)
        return False
    present = {row.get("name") for row in rows}
    missing = [table for table in tables if table not in present]
    if missing:
        logger.info("Sankey tables missing: %s", ", ".join(missing))
        return False
    return True


async def _columns_present(
    client: Any,
    table: str,
    columns: List[str],
) -> bool:
    if not columns:
        return True
    try:
        rows = await query_dicts(
            client,
            """
            SELECT name
            FROM system.columns
            WHERE database = currentDatabase()
                AND table = %(table)s
                AND name IN %(columns)s
            """,
            {"table": table, "columns": columns},
        )
    except Exception as exc:
        logger.warning("Sankey column lookup failed for %s: %s", table, exc)
        return False
    present = {row.get("name") for row in rows}
    missing = [column for column in columns if column not in present]
    if missing:
        logger.info("Sankey columns missing for %s: %s", table, ", ".join(missing))
        return False
    return True


async def _repo_scope_filter(
    client: Any,
    filters: MetricFilter,
    repo_column: str = "repo_id",
) -> Tuple[str, Dict[str, Any]]:
    repo_ids = await resolve_repo_filter_ids(client, filters)
    if not repo_ids:
        return "", {}
    return build_scope_filter_multi("repo", repo_ids, repo_column=repo_column)


def _team_scope_filter(
    filters: MetricFilter,
    team_column: str = "team_id",
) -> Tuple[str, Dict[str, Any]]:
    if filters.scope.level != "team" or not filters.scope.ids:
        return "", {}
    return build_scope_filter_multi("team", filters.scope.ids, team_column=team_column)


def _category_theme_filters(filters: MetricFilter) -> List[str]:
    raw_categories = filters.why.work_category or []
    themes: List[str] = []
    for category in raw_categories:
        if not category:
            continue
        category_str = str(category).strip()
        if not category_str:
            continue
        if "." in category_str:
            theme = category_str.split(".", 1)[0]
        else:
            theme = category_str
        if theme:
            themes.append(theme)
    return list(dict.fromkeys(themes))


def _work_scope_filter(
    filters: MetricFilter,
    work_scope_column: str = "work_scope_id",
) -> Tuple[str, Dict[str, Any]]:
    scope_ids: List[str] = []
    if filters.scope.level == "repo":
        scope_ids.extend(filters.scope.ids)
    if filters.what.repos:
        scope_ids.extend(filters.what.repos)
    scope_ids = [scope_id for scope_id in scope_ids if scope_id]
    if not scope_ids:
        return "", {}
    return f" AND {work_scope_column} IN %(work_scope_ids)s", {
        "work_scope_ids": scope_ids
    }


async def _build_investment_flow(
    client: Any,
    *,
    start_day: date,
    end_day: date,
    filters: MetricFilter,
) -> Tuple[List[SankeyNode], List[SankeyLink]]:
    if not await _tables_present(client, ["work_unit_investments"]):
        return [], []
    if not await _columns_present(
        client,
        "work_unit_investments",
        [
            "theme_distribution_json",
            "effort_value",
            "from_ts",
            "to_ts",
            "repo_id",
        ],
    ):
        return [], []
    repo_filter, repo_params = await _repo_scope_filter(
        client, filters, repo_column="repo_id"
    )
    theme_filters = _category_theme_filters(filters)
    category_filter = " AND theme_kv.1 IN %(themes)s" if theme_filters else ""
    scope_filter = f"{repo_filter}{category_filter}"
    scope_params = {**repo_params}
    if theme_filters:
        scope_params["themes"] = theme_filters
    window_start = datetime.combine(start_day, time.min, tzinfo=timezone.utc)
    window_end = datetime.combine(end_day, time.min, tzinfo=timezone.utc)
    rows = await fetch_investment_flow_items(
        client,
        start_ts=window_start,
        end_ts=window_end,
        scope_filter=scope_filter,
        scope_params=scope_params,
        limit=MAX_INVESTMENT_ITEMS,
    )
    nodes: Dict[str, SankeyNode] = {}
    edges: Dict[Tuple[str, str], float] = {}
    for row in rows:
        value = float(row.get("value") or 0.0)
        if value <= 0:
            continue
        source = _normalize_label(row.get("source"), "Unassigned")
        target = _normalize_label(row.get("target"), "Other")

        _touch_node(nodes, source, "initiative")
        _touch_node(nodes, target, "project")

        _add_edge(edges, source, target, value)

    return list(nodes.values()), _links_from_edges(edges)


async def _build_expense_flow(
    client: Any,
    *,
    start_day: date,
    end_day: date,
    filters: MetricFilter,
) -> Tuple[List[SankeyNode], List[SankeyLink]]:
    if not await _tables_present(
        client, ["work_item_metrics_daily", "work_item_cycle_times"]
    ):
        return [], []
    if not await _columns_present(
        client,
        "work_item_metrics_daily",
        [
            "day",
            "new_items_count",
            "new_bugs_count",
            "items_completed",
            "bug_completed_ratio",
            "team_id",
            "work_scope_id",
        ],
    ):
        return [], []
    if not await _columns_present(
        client,
        "work_item_cycle_times",
        ["day", "status", "team_id", "work_scope_id"],
    ):
        return [], []
    team_filter, team_params = _team_scope_filter(
        filters, team_column="ifNull(nullIf(team_id, ''), 'unassigned')"
    )
    work_scope_filter, work_scope_params = _work_scope_filter(
        filters, work_scope_column="work_scope_id"
    )
    scope_filter = f"{team_filter}{work_scope_filter}"
    scope_params = {**team_params, **work_scope_params}
    rows = await fetch_expense_counts(
        client,
        start_day=start_day,
        end_day=end_day,
        scope_filter=scope_filter,
        scope_params=scope_params,
    )
    if not rows:
        return [], []
    row = rows[0]
    new_bugs = float(row.get("new_bugs") or 0.0)
    bug_completed = float(row.get("bug_completed_estimate") or 0.0)

    abandoned_rows = await fetch_expense_abandoned(
        client,
        start_day=start_day,
        end_day=end_day,
        scope_filter=scope_filter,
        scope_params=scope_params,
    )
    canceled_items = 0.0
    if abandoned_rows:
        canceled_items = float(abandoned_rows[0].get("canceled_items") or 0.0)

    unplanned = max(0.0, new_bugs)
    rework = max(0.0, min(unplanned, bug_completed))
    abandoned = max(0.0, min(rework, canceled_items))

    nodes: Dict[str, SankeyNode] = {}
    edges: Dict[Tuple[str, str], float] = {}

    _touch_node(nodes, "Planned work", "planned")
    _touch_node(nodes, "Unplanned work", "unplanned")
    _touch_node(nodes, "Rework", "rework")
    _touch_node(nodes, "Abandonment / rewrite", "abandonment")

    _add_edge(edges, "Planned work", "Unplanned work", unplanned)
    _add_edge(edges, "Unplanned work", "Rework", rework)
    _add_edge(edges, "Rework", "Abandonment / rewrite", abandoned)

    return list(nodes.values()), _links_from_edges(edges)


async def _build_state_flow(
    client: Any,
    *,
    start_day: date,
    end_day: date,
    filters: MetricFilter,
) -> Tuple[List[SankeyNode], List[SankeyLink]]:
    if not await _tables_present(client, ["work_item_state_durations_daily"]):
        return [], []
    if not await _columns_present(
        client,
        "work_item_state_durations_daily",
        ["day", "status", "items_touched", "team_id", "work_scope_id"],
    ):
        return [], []
    team_filter, team_params = _team_scope_filter(
        filters, team_column="ifNull(nullIf(team_id, ''), 'unassigned')"
    )
    work_scope_filter, work_scope_params = _work_scope_filter(
        filters, work_scope_column="work_scope_id"
    )
    scope_filter = f"{team_filter}{work_scope_filter}"
    scope_params = {**team_params, **work_scope_params}
    rows = await fetch_state_status_counts(
        client,
        start_day=start_day,
        end_day=end_day,
        scope_filter=scope_filter,
        scope_params=scope_params,
    )
    nodes: Dict[str, SankeyNode] = {}
    edges: Dict[Tuple[str, str], float] = {}

    status_counts: Dict[str, float] = {}
    for row in rows:
        value = float(row.get("items_touched") or 0.0)
        if value <= 0:
            continue
        status_raw = _normalize_label(row.get("status"), "unknown").lower()
        status_counts[status_raw] = status_counts.get(status_raw, 0.0) + value

    status_labels = {
        "backlog": "Backlog",
        "todo": "Todo",
        "in_progress": "In Progress",
        "in_review": "In Review",
        "blocked": "Blocked",
        "done": "Done",
        "canceled": "Canceled",
        "unknown": "Unknown",
    }

    def _label(status: str) -> str:
        return status_labels.get(status, status.replace("_", " ").title())

    backlog = status_counts.get("backlog", 0.0)
    todo = status_counts.get("todo", 0.0)
    in_progress = status_counts.get("in_progress", 0.0)
    in_review = status_counts.get("in_review", 0.0)
    done = status_counts.get("done", 0.0)
    blocked = status_counts.get("blocked", 0.0)
    canceled = status_counts.get("canceled", 0.0)

    flow_backlog = min(backlog, todo)
    flow_todo = min(todo, in_progress)
    blocked_flow = min(in_progress, blocked)
    remaining = max(0.0, in_progress - blocked_flow)
    review_flow = min(remaining, in_review)
    remaining = max(0.0, remaining - review_flow)
    canceled_flow = min(remaining, canceled)
    done_flow = min(review_flow, done)

    for status in status_counts:
        _touch_node(nodes, _label(status), "state")

    _add_edge(edges, _label("backlog"), _label("todo"), flow_backlog)
    _add_edge(edges, _label("todo"), _label("in_progress"), flow_todo)
    _add_edge(edges, _label("in_progress"), _label("blocked"), blocked_flow)
    _add_edge(edges, _label("in_progress"), _label("in_review"), review_flow)
    _add_edge(edges, _label("in_progress"), _label("canceled"), canceled_flow)
    _add_edge(edges, _label("in_review"), _label("done"), done_flow)

    return list(nodes.values()), _links_from_edges(edges)


async def _build_hotspot_flow(
    client: Any,
    *,
    start_day: date,
    end_day: date,
    filters: MetricFilter,
) -> Tuple[List[SankeyNode], List[SankeyLink]]:
    if not await _tables_present(client, ["file_metrics_daily", "repos"]):
        return [], []
    if not await _columns_present(
        client,
        "file_metrics_daily",
        ["repo_id", "day", "path", "churn"],
    ):
        return [], []
    if not await _columns_present(client, "repos", ["id", "repo"]):
        return [], []
    scope_filter, scope_params = await _repo_scope_filter(
        client, filters, repo_column="metrics.repo_id"
    )
    rows = await fetch_hotspot_rows(
        client,
        start_day=start_day,
        end_day=end_day,
        scope_filter=scope_filter,
        scope_params=scope_params,
        limit=MAX_HOTSPOT_ROWS,
    )
    nodes: Dict[str, SankeyNode] = {}
    edges: Dict[Tuple[str, str], float] = {}
    for row in rows:
        churn = float(row.get("churn") or 0.0)
        if churn <= 0:
            continue
        repo = _normalize_label(row.get("repo"), "Unknown repo")
        directory = _normalize_label(row.get("directory"), "(root)")
        file_path = _normalize_label(row.get("file_path"), "unknown file")
        change_type = _normalize_label(row.get("change_type"), "feature")

        directory_label = f"{repo} / {directory}"
        file_label = f"{repo} / {file_path}"

        _touch_node(nodes, repo, "repo")
        _touch_node(nodes, directory_label, "directory")
        _touch_node(nodes, file_label, "file")
        _touch_node(nodes, change_type, "change_type")

        _add_edge(edges, repo, directory_label, churn)
        _add_edge(edges, directory_label, file_label, churn)
        _add_edge(edges, file_label, change_type, churn)

    return list(nodes.values()), _links_from_edges(edges)


async def build_sankey_response(
    *,
    db_url: str,
    mode: str,
    filters: MetricFilter,
    context: Optional[SankeyContext] = None,
    window_start: Optional[date] = None,
    window_end: Optional[date] = None,
) -> SankeyResponse:
    if mode == "hotpot":
        mode = "hotspot"
    definition = SANKEY_DEFINITIONS.get(mode)
    if definition is None:
        raise ValueError(f"Unknown sankey mode: {mode}")

    resolved_filters = _apply_window_to_filters(filters, window_start, window_end)
    start_day, end_day, _, _ = time_window(resolved_filters)

    async with clickhouse_client(db_url) as client:
        if mode == "investment":
            nodes, links = await _build_investment_flow(
                client,
                start_day=start_day,
                end_day=end_day,
                filters=resolved_filters,
            )
        elif mode == "expense":
            nodes, links = await _build_expense_flow(
                client,
                start_day=start_day,
                end_day=end_day,
                filters=resolved_filters,
            )
        elif mode == "state":
            nodes, links = await _build_state_flow(
                client,
                start_day=start_day,
                end_day=end_day,
                filters=resolved_filters,
            )
        elif mode == "hotspot":
            nodes, links = await _build_hotspot_flow(
                client,
                start_day=start_day,
                end_day=end_day,
                filters=resolved_filters,
            )
        else:
            nodes, links = [], []

    return SankeyResponse(
        mode=mode,
        nodes=nodes,
        links=links,
        unit=definition.unit,
        label=definition.label,
        description=definition.description,
    )



import pytest
import uuid
from datetime import datetime, timezone, timedelta
from unittest.mock import MagicMock
from dev_health_ops.metrics.loaders.clickhouse import ClickHouseDataLoader
from dev_health_ops.metrics.loaders.mongo import MongoDataLoader
from dev_health_ops.metrics.loaders.sqlalchemy import SqlAlchemyDataLoader
from sqlalchemy import create_engine


@pytest.mark.asyncio
async def test_clickhouse_loader_async():
    mock_client = MagicMock()
    # Mocking successful queries
    mock_result = MagicMock()
    mock_result.column_names = [
        "repo_id",
        "commit_hash",
        "author_email",
        "author_name",
        "committer_when",
        "file_path",
        "additions",
        "deletions",
    ]
    mock_result.result_rows = []
    mock_client.query.return_value = mock_result

    loader = ClickHouseDataLoader(mock_client)
    start = datetime.now(timezone.utc)
    end = start + timedelta(days=1)

    # Verify it's awaitable and doesn't crash
    res = await loader.load_git_rows(start, end, uuid.uuid4())
    assert len(res) == 3
    assert isinstance(res[0], list)


@pytest.mark.asyncio
async def test_mongo_loader_async():
    mock_db = MagicMock()
    loader = MongoDataLoader(mock_db)
    start = datetime.now(timezone.utc)
    end = start + timedelta(days=1)

    # Mock some mongo responses
    mock_db["git_commits"].find.return_value = []
    mock_db["git_commit_stats"].find.return_value = []
    mock_db["git_pull_requests"].find.return_value = []
    mock_db["git_pull_request_reviews"].find.return_value = []

    res = await loader.load_git_rows(start, end, uuid.uuid4())
    assert len(res) == 3


@pytest.mark.asyncio
async def test_sqlalchemy_loader_async():
    engine = create_engine("sqlite:///:memory:")
    # Simple setup: tables usually exist, but we just verify the async call wrap
    loader = SqlAlchemyDataLoader(engine)
    start = datetime.now(timezone.utc)
    end = start + timedelta(days=1)

    # This will likely fail on missing tables if executed, but we want to verify the async signature
    with pytest.raises(Exception):
        await loader.load_git_rows(start, end, uuid.uuid4())



from __future__ import annotations

import json
from datetime import datetime, time, timezone
from typing import Dict, List, Optional, Tuple

from dev_health_ops.investment_taxonomy import SUBCATEGORIES, THEMES

from ..models.filters import MetricFilter
from ..models.schemas import (
    EvidenceQuality,
    InvestmentBreakdown,
    WorkUnitEvidence,
    WorkUnitEffort,
    WorkUnitInvestment,
    WorkUnitTimeRange,
)
from ..queries.client import clickhouse_client
from ..queries.work_unit_investments import (
    fetch_work_unit_investment_quotes,
    fetch_work_unit_investments,
)
from .filtering import resolve_repo_filter_ids, time_window


def _normalize_scores(scores: Dict[str, float], keys: List[str]) -> Dict[str, float]:
    total = sum(float(scores.get(key, 0.0) or 0.0) for key in keys)
    if total <= 0.0:
        uniform = 1.0 / len(keys) if keys else 0.0
        return {key: uniform for key in keys}
    return {key: float(scores.get(key, 0.0) or 0.0) / total for key in keys}


def _evidence_quality_band(value: float) -> str:
    if value >= 0.8:
        return "high"
    if value >= 0.6:
        return "moderate"
    if value >= 0.4:
        return "low"
    return "very_low"


def _parse_distribution(value: object) -> Dict[str, float]:
    if isinstance(value, dict):
        return {str(k): float(v or 0.0) for k, v in value.items()}
    if isinstance(value, str):
        try:
            parsed = json.loads(value)
        except json.JSONDecodeError:
            return {}
        if isinstance(parsed, dict):
            return {str(k): float(v or 0.0) for k, v in parsed.items()}
    return {}


def _segment_filter_match(
    theme_distribution: Dict[str, float],
    subcategory_distribution: Dict[str, float],
    *,
    theme: Optional[str],
    subcategory: Optional[str],
) -> bool:
    if subcategory:
        return float(subcategory_distribution.get(subcategory, 0.0)) > 0.0
    if theme:
        return float(theme_distribution.get(theme, 0.0)) > 0.0
    return True


def _segment_contribution(
    theme_distribution: Dict[str, float],
    subcategory_distribution: Dict[str, float],
    *,
    theme: Optional[str],
    subcategory: Optional[str],
    effort_value: float,
) -> float:
    if subcategory:
        return effort_value * float(subcategory_distribution.get(subcategory, 0.0))
    if theme:
        return effort_value * float(theme_distribution.get(theme, 0.0))
    return effort_value


async def build_segment_investment(
    *,
    db_url: str,
    filters: MetricFilter,
    theme: Optional[str],
    subcategory: Optional[str],
    limit: int = 500,
) -> Optional[WorkUnitInvestment]:
    start_day, end_day, _, _ = time_window(filters)
    start_ts = datetime.combine(start_day, time.min, tzinfo=timezone.utc)
    end_ts = datetime.combine(end_day, time.min, tzinfo=timezone.utc)

    async with clickhouse_client(db_url) as client:
        repo_ids = await resolve_repo_filter_ids(client, filters)
        rows = await fetch_work_unit_investments(
            client,
            start_ts=start_ts,
            end_ts=end_ts,
            repo_ids=repo_ids or None,
            limit=max(1, int(limit)),
        )

        if not rows:
            return None

        filtered_rows = []
        for row in rows:
            theme_distribution = _parse_distribution(row.get("theme_distribution_json"))
            subcategory_distribution = _parse_distribution(
                row.get("subcategory_distribution_json")
            )
            if _segment_filter_match(
                theme_distribution,
                subcategory_distribution,
                theme=theme,
                subcategory=subcategory,
            ):
                filtered_rows.append(row)
        rows = filtered_rows

        if not rows:
            return None

        total_effort = 0.0
        theme_totals: Dict[str, float] = {key: 0.0 for key in THEMES}
        subcategory_totals: Dict[str, float] = {key: 0.0 for key in SUBCATEGORIES}
        evidence_weighted = 0.0
        contributions: List[Tuple[float, str, str]] = []

        for row in rows:
            effort_value = float(row.get("effort_value") or 0.0)
            theme_distribution = _parse_distribution(row.get("theme_distribution_json"))
            subcategory_distribution = _parse_distribution(
                row.get("subcategory_distribution_json")
            )
            total_effort += effort_value
            evidence_weighted += effort_value * float(
                row.get("evidence_quality") or 0.0
            )
            for key, value in theme_distribution.items():
                if key in theme_totals:
                    theme_totals[key] += effort_value * float(value or 0.0)
            for key, value in subcategory_distribution.items():
                if key in subcategory_totals:
                    subcategory_totals[key] += effort_value * float(value or 0.0)

            contribution = _segment_contribution(
                theme_distribution,
                subcategory_distribution,
                theme=theme,
                subcategory=subcategory,
                effort_value=effort_value,
            )
            if contribution > 0:
                contributions.append(
                    (
                        contribution,
                        str(row.get("work_unit_id") or ""),
                        str(row.get("categorization_run_id") or ""),
                    )
                )

        theme_distribution = _normalize_scores(theme_totals, sorted(THEMES))
        subcategory_distribution = _normalize_scores(
            subcategory_totals, sorted(SUBCATEGORIES)
        )

        evidence_quality_value = (
            evidence_weighted / total_effort if total_effort > 0 else 0.0
        )
        evidence_quality = EvidenceQuality(
            value=evidence_quality_value,
            band=_evidence_quality_band(evidence_quality_value),
        )

        contributions.sort(key=lambda item: item[0], reverse=True)
        unit_runs = [
            (unit_id, run_id)
            for _, unit_id, run_id in contributions[:5]
            if unit_id and run_id
        ]
        quote_rows = await fetch_work_unit_investment_quotes(
            client, unit_runs=unit_runs
        )

    textual_evidence: List[Dict[str, object]] = []
    for quote in quote_rows or []:
        textual_evidence.append(
            {
                "type": "evidence_quote",
                "quote": quote.get("quote"),
                "source": quote.get("source_type"),
                "id": quote.get("source_id"),
            }
        )

    structural_evidence: List[Dict[str, object]] = []
    if contributions:
        structural_evidence.append(
            {
                "type": "segment_work_units",
                "work_unit_ids": [unit_id for _, unit_id, _ in contributions[:10]],
            }
        )

    segment_id = "segment:all"
    if subcategory:
        segment_id = f"segment:{subcategory}"
    elif theme:
        segment_id = f"segment:{theme}"

    return WorkUnitInvestment(
        work_unit_id=segment_id,
        time_range=WorkUnitTimeRange(start=start_ts, end=end_ts),
        effort=WorkUnitEffort(metric="churn_loc", value=total_effort),
        investment=InvestmentBreakdown(
            themes=theme_distribution,
            subcategories=subcategory_distribution,
        ),
        evidence_quality=evidence_quality,
        evidence=WorkUnitEvidence(
            textual=textual_evidence,
            structural=structural_evidence,
            contextual=[
                {
                    "type": "time_range",
                    "start": start_ts.isoformat(),
                    "end": end_ts.isoformat(),
                }
            ],
        ),
    )



from __future__ import annotations

import pytest
import json
from unittest.mock import AsyncMock, MagicMock
from dev_health_ops.llm.providers.openai import OpenAIProvider


@pytest.mark.asyncio
async def test_gpt5_truncation_retry_logic():
    # Setup facade with GPT-5 model
    provider = OpenAIProvider(
        api_key="test", model="gpt-5-mini", max_completion_tokens=1024
    )

    mock_client = AsyncMock()

    # First response: truncated
    resp1 = MagicMock()
    resp1.output_text = '{"summary": "This is truncated'
    resp1.incomplete_details = MagicMock(reason="max_output_tokens")

    # Second response: valid and complete
    resp2 = MagicMock()
    resp2.output_text = '{"summary": "This is complete now"}'
    resp2.incomplete_details = None

    mock_client.responses.create.side_effect = [resp1, resp2]

    # Inject mock client into implementation
    provider._impl._client = mock_client

    result_json = await provider.complete("explain something")

    # Verify parsing
    result = json.loads(result_json)
    assert result["summary"] == "This is complete now"

    # Verify two calls
    assert mock_client.responses.create.call_count == 2

    # Verify token budget increase (First call default for explanation is 2048)
    first_call_kwargs = mock_client.responses.create.call_args_list[0].kwargs
    second_call_kwargs = mock_client.responses.create.call_args_list[1].kwargs

    assert first_call_kwargs["max_output_tokens"] == 4096
    assert second_call_kwargs["max_output_tokens"] == 8192

    # Verify validate_json_or_empty output is compact (json.dumps)
    assert result_json == json.dumps(result)



"""Authorization utilities for GraphQL analytics API."""

from __future__ import annotations

from typing import Any, Dict

from .context import GraphQLContext
from .errors import AuthorizationError


def require_org_id(context: GraphQLContext) -> str:
    """
    Validate that org_id is present in the context.

    Args:
        context: GraphQL request context.

    Returns:
        The validated org_id.

    Raises:
        AuthorizationError: If org_id is missing or empty.
    """
    if not context.org_id:
        raise AuthorizationError("org_id is required for all analytics queries")
    return context.org_id


def enforce_org_scope(org_id: str, params: Dict[str, Any]) -> Dict[str, Any]:
    """
    Inject org_id into SQL parameters to enforce org scoping.

    This function ensures that org_id is always included in query parameters,
    which is then used in WHERE clauses to scope data access.

    Args:
        org_id: The organization ID to scope queries to.
        params: Existing SQL parameters dict.

    Returns:
        Updated params dict with org_id included.

    Raises:
        AuthorizationError: If org_id is missing or empty.
    """
    if not org_id:
        raise AuthorizationError("org_id is required for org scoping")

    # Create a new dict to avoid mutating the original
    scoped_params = dict(params)
    scoped_params["org_id"] = org_id
    return scoped_params


def validate_org_access(context: GraphQLContext, requested_org_id: str) -> None:
    """
    Validate that the context org_id matches the requested org_id.

    This prevents cross-org data access by ensuring the authenticated
    org matches the requested data scope.

    Args:
        context: GraphQL request context.
        requested_org_id: The org_id being requested in the query.

    Raises:
        AuthorizationError: If org_ids don't match.
    """
    if context.org_id != requested_org_id:
        raise AuthorizationError(
            f"Access denied: cannot query data for org '{requested_org_id}'"
        )



from dev_health_ops.audit.perf import format_perf_report, run_perf_audit


def test_perf_audit_unsupported_backend() -> None:
    report = run_perf_audit("sqlite:///tmp/test.db")
    assert report["status"] == "unchecked"
    assert report["db_error"] == "unsupported backend: sqlite"


def test_format_perf_report_unchecked() -> None:
    report = {
        "status": "unchecked",
        "db_backend": "clickhouse",
        "db_error": "connection failed",
        "threshold_ms": 1000,
        "lookback_minutes": 60,
        "limit": 20,
        "log_settings": {},
        "slow_queries": [],
        "notes": [],
    }
    output = format_perf_report(report)
    assert "status: unchecked" in output
    assert "db_error: connection failed" in output



import asyncio
import uuid

import pytest

from dev_health_ops.processors.github import _sync_github_prs_to_store
from dev_health_ops.processors.gitlab import _sync_gitlab_mrs_to_store


class _NoSleepGate:
    def __init__(self):
        self.penalties = []

    def wait_sync(self) -> None:
        return

    def penalize(self, delay_seconds=None) -> float:
        self.penalties.append(delay_seconds)
        return float(delay_seconds or 0)

    def reset(self) -> None:
        return


class _FakeStore:
    def __init__(self):
        self.pr_batches = []

    async def insert_git_pull_requests(self, batch):
        # store a shallow copy to avoid later mutation surprises
        self.pr_batches.append(list(batch))


class _FakeGithub:
    def __init__(self, repo):
        self._repo = repo

    def get_repo(self, _full_name: str):
        return self._repo


class _RetryAfterException(Exception):
    def __init__(self, headers):
        super().__init__("rate limited")
        self.headers = headers


class _PRIter:
    def __init__(self, items):
        self._items = list(items)
        self._idx = 0

    def __iter__(self):
        return self

    def __next__(self):
        if self._idx >= len(self._items):
            raise StopIteration
        item = self._items[self._idx]
        self._idx += 1
        if isinstance(item, Exception):
            raise item
        return item


class _FakePRUser:
    def __init__(self, login="octo", email=None):
        self.login = login
        self.email = email


class _FakePRRef:
    def __init__(self, ref):
        self.ref = ref


class _FakePR:
    def __init__(self, number: int):
        self.number = number
        self.title = f"PR {number}"
        self.state = "closed"
        self.user = _FakePRUser()
        self.created_at = None
        self.merged_at = None
        self.closed_at = None
        self.head = _FakePRRef("feature")
        self.base = _FakePRRef("main")


class _FakeRepo:
    def __init__(self, pull_items):
        self._pull_items = pull_items

    def get_pulls(self, state="all"):
        assert state == "all"
        return _PRIter(self._pull_items)


@pytest.mark.asyncio
async def test_github_pr_sync_retries_on_retry_after_and_persists():
    loop = asyncio.get_running_loop()
    repo_id = uuid.uuid4()
    store = _FakeStore()

    # First call hits a Retry-After style limit, then we yield two PRs.
    fake_repo = _FakeRepo(
        [
            _RetryAfterException({"Retry-After": "0"}),
            _FakePR(1),
            _FakePR(2),
        ]
    )

    class _Connector:
        def __init__(self):
            self.github = _FakeGithub(fake_repo)

    connector = _Connector()
    gate = _NoSleepGate()

    total = await loop.run_in_executor(
        None,
        _sync_github_prs_to_store,
        connector,
        "o",
        "r",
        repo_id,
        store,
        loop,
        1,  # batch_size
        "all",
        gate,
    )

    assert total == 2
    assert len(store.pr_batches) == 2
    assert [b[0].number for b in store.pr_batches] == [1, 2]
    assert gate.penalties and gate.penalties[0] == pytest.approx(0.0)


@pytest.mark.asyncio
async def test_gitlab_mr_sync_retries_on_retry_after_and_persists():
    loop = asyncio.get_running_loop()
    repo_id = uuid.uuid4()
    store = _FakeStore()

    class _RateLimit(Exception):
        def __init__(self, retry_after_seconds):
            super().__init__("rate limited")
            self.retry_after_seconds = retry_after_seconds

    class _Rest:
        def __init__(self):
            self.calls = 0

        def get_merge_requests(
            self,
            project_id=None,
            state=None,
            page=None,
            per_page=None,
            **_kwargs,
        ):
            assert project_id is not None
            assert page is not None
            assert per_page is not None
            assert state == "all"
            self.calls += 1
            if self.calls == 1:
                raise _RateLimit(0)
            if self.calls == 2:
                return [
                    {
                        "iid": 7,
                        "title": "MR 7",
                        "state": "merged",
                        "created_at": "2020-01-01T00:00:00Z",
                        "merged_at": None,
                        "closed_at": None,
                        "source_branch": "feature",
                        "target_branch": "main",
                        "author": {"username": "alice"},
                    }
                ]
            return []

    class _Connector:
        def __init__(self):
            self.per_page = 100
            self.rest_client = _Rest()

    connector = _Connector()
    gate = _NoSleepGate()

    total = await loop.run_in_executor(
        None,
        _sync_gitlab_mrs_to_store,
        connector,
        123,
        repo_id,
        store,
        loop,
        1,  # batch_size
        "all",
        gate,
    )

    assert total == 1
    assert len(store.pr_batches) == 1
    assert store.pr_batches[0][0].number == 7
    assert gate.penalties and gate.penalties[0] == 0.0



"""
Extractors for parsing links from various sources.
"""

from dev_health_ops.work_graph.extractors.text_parser import (
    extract_jira_keys,
    extract_github_issue_refs,
    extract_gitlab_issue_refs,
    ParsedIssueRef,
)

__all__ = [
    "extract_jira_keys",
    "extract_github_issue_refs",
    "extract_gitlab_issue_refs",
    "ParsedIssueRef",
]



from __future__ import annotations

import argparse
import json
import re
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Sequence

from dev_health_ops.metrics.sinks.clickhouse import ClickHouseMetricsSink
from dev_health_ops.storage import detect_db_type

AUDIT_PROVIDERS = ("jira", "github", "gitlab", "synthetic")

REQUIRED_CONFIG_KEYS: Dict[str, Sequence[str]] = {
    "jira": ("JIRA_BASE_URL", "JIRA_EMAIL", "JIRA_API_TOKEN"),
    "github": ("GITHUB_TOKEN",),
    "gitlab": ("GITLAB_TOKEN",),
    "synthetic": (),
}

COLLECTOR_SPECS: Dict[str, Dict[str, Sequence[str]]] = {
    "jira": {
        "modules": ("providers/jira/client.py",),
        "entrypoints": ("fetch_jira_work_items_with_extras", "fetch_jira_work_items"),
    },
    "github": {
        "modules": ("providers/github/client.py",),
        "entrypoints": ("fetch_github_work_items", "fetch_github_project_v2_items"),
    },
    "gitlab": {
        "modules": ("providers/gitlab/client.py",),
        "entrypoints": ("fetch_gitlab_work_items",),
    },
    "synthetic": {
        "modules": ("metrics/work_items.py",),
        "entrypoints": ("fetch_synthetic_work_items",),
    },
}

REQUIRED_TABLES = ("work_items", "work_item_transitions")
REQUIRED_COLUMNS = {
    "work_items": {
        "repo_id",
        "work_item_id",
        "provider",
        "title",
        "type",
        "status",
        "status_raw",
        "project_key",
        "project_id",
        "assignees",
        "reporter",
        "created_at",
        "updated_at",
        "started_at",
        "completed_at",
        "closed_at",
        "labels",
        "story_points",
        "sprint_id",
        "sprint_name",
        "parent_id",
        "epic_id",
        "url",
        "last_synced",
    },
    "work_item_transitions": {
        "repo_id",
        "work_item_id",
        "occurred_at",
        "provider",
        "from_status",
        "to_status",
        "from_status_raw",
        "to_status_raw",
        "actor",
        "last_synced",
    },
}


def parse_provider_list(raw: Optional[str]) -> List[str]:
    if raw is None or not str(raw).strip():
        return list(AUDIT_PROVIDERS)
    normalized = str(raw).strip().lower()
    if normalized in {"all", "*"}:
        return list(AUDIT_PROVIDERS)
    parts = [p.strip().lower() for p in normalized.split(",") if p.strip()]
    unknown = [p for p in parts if p not in AUDIT_PROVIDERS]
    if unknown:
        raise ValueError(f"Unknown provider(s): {', '.join(sorted(set(unknown)))}")
    return parts


def _read_text(path: Path) -> str:
    try:
        return path.read_text(encoding="utf-8", errors="ignore")
    except FileNotFoundError:
        return ""


def _has_def(text: str, name: str) -> bool:
    return re.search(rf"\bdef\s+{re.escape(name)}\b", text) is not None


def _find_in_files(token: str, files: Iterable[Path]) -> bool:
    for path in files:
        if token in _read_text(path):
            return True
    return False


def _list_files(root: Path, suffix: str) -> List[Path]:
    if not root.exists():
        return []
    return [p for p in root.rglob(f"*{suffix}") if p.is_file()]


def _collector_check(repo_root: Path) -> Dict[str, Dict[str, Any]]:
    work_items_text = _read_text(repo_root / "metrics/work_items.py")
    results: Dict[str, Dict[str, Any]] = {}
    for provider, spec in COLLECTOR_SPECS.items():
        modules = [repo_root / path for path in spec.get("modules", [])]
        missing_modules = [str(path) for path in modules if not path.exists()]
        entrypoints = list(spec.get("entrypoints", []))
        entrypoint_ok = any(_has_def(work_items_text, name) for name in entrypoints)
        missing_entrypoints = [] if entrypoint_ok else entrypoints
        ok = not missing_modules and entrypoint_ok
        results[provider] = {
            "status": "ok" if ok else "missing",
            "missing_modules": missing_modules,
            "missing_entrypoints": missing_entrypoints,
        }
    return results


def _config_check(repo_root: Path) -> Dict[str, Dict[str, Any]]:
    code_files = (
        _list_files(repo_root / "providers", ".py")
        + _list_files(repo_root / "metrics", ".py")
        + [repo_root / "cli.py"]
    )
    doc_files = [repo_root / "README.md"] + _list_files(repo_root / "docs", ".md")
    results: Dict[str, Dict[str, Any]] = {}
    for provider, keys in REQUIRED_CONFIG_KEYS.items():
        missing_code = [k for k in keys if not _find_in_files(k, code_files)]
        missing_docs = [k for k in keys if not _find_in_files(k, doc_files)]
        ok = not missing_code and not missing_docs
        results[provider] = {
            "status": "ok" if ok else "missing",
            "missing_in_code": missing_code,
            "missing_in_docs": missing_docs,
        }
    return results


def _sink_check(repo_root: Path) -> Dict[str, Dict[str, Any]]:
    sink_text = _read_text(repo_root / "metrics/sinks/clickhouse.py")
    writers_ok = all(
        _has_def(sink_text, name)
        for name in ("write_work_items", "write_work_item_transitions")
    )
    missing_writers = []
    if not writers_ok:
        missing_writers = ["write_work_items", "write_work_item_transitions"]

    job_text = _read_text(repo_root / "metrics/job_work_items.py")
    results: Dict[str, Dict[str, Any]] = {}
    for provider in AUDIT_PROVIDERS:
        mapping_ok = provider in job_text
        missing = []
        if missing_writers:
            missing.extend(missing_writers)
        if not mapping_ok:
            missing.append("provider_mapping")
        ok = writers_ok and mapping_ok
        results[provider] = {
            "status": "ok" if ok else "missing",
            "missing": missing,
        }
    return results


def _extract_work_item_provider_choices(cli_text: str) -> List[str]:
    match = re.search(
        r"wi\.add_argument\([\s\S]*?--provider[\s\S]*?choices=\[(.*?)\]",
        cli_text,
    )
    if not match:
        return []
    choices = match.group(1)
    return re.findall(r"[\"']([a-zA-Z0-9_]+)[\"']", choices)


def _commands_check(repo_root: Path) -> Dict[str, Dict[str, Any]]:
    cli_text = _read_text(repo_root / "cli.py")
    provider_choices = set(_extract_work_item_provider_choices(cli_text))
    job_text = _read_text(repo_root / "metrics/job_work_items.py")
    job_entrypoint_ok = _has_def(job_text, "run_work_items_sync_job")

    results: Dict[str, Dict[str, Any]] = {}
    for provider in AUDIT_PROVIDERS:
        cli_ok = provider in provider_choices
        job_ok = job_entrypoint_ok and provider in job_text
        ok = cli_ok or job_ok
        missing = []
        if not ok:
            missing.append("sync_work_items_command")
        results[provider] = {
            "status": "ok" if ok else "missing",
            "missing": missing,
        }
    return results


def _migration_tables(repo_root: Path, tables: Sequence[str]) -> List[str]:
    migrations_dir = repo_root / "migrations/clickhouse"
    if not migrations_dir.exists():
        return list(tables)
    files = _list_files(migrations_dir, ".sql")
    found: Dict[str, bool] = {table: False for table in tables}
    patterns = {
        table: re.compile(
            rf"CREATE\s+TABLE\s+(IF\s+NOT\s+EXISTS\s+)?{re.escape(table)}\b",
            re.IGNORECASE,
        )
        for table in tables
    }
    for path in files:
        text = _read_text(path)
        for table, pattern in patterns.items():
            if found[table]:
                continue
            if pattern.search(text):
                found[table] = True
    return [table for table, present in found.items() if not present]


def _query_dicts(
    client: Any, query: str, parameters: Dict[str, Any]
) -> List[Dict[str, Any]]:
    result = client.query(query, parameters=parameters)
    col_names = list(getattr(result, "column_names", []) or [])
    rows = list(getattr(result, "result_rows", []) or [])
    if not col_names or not rows:
        return []
    return [dict(zip(col_names, row)) for row in rows]


def _schema_check(db_url: str, repo_root: Path) -> Dict[str, Any]:
    missing_migrations = _migration_tables(repo_root, REQUIRED_TABLES)
    info: Dict[str, Any] = {
        "status": "unchecked",
        "checked": False,
        "missing_tables": [],
        "missing_columns": {},
        "missing_migrations": missing_migrations,
        "db_backend": None,
        "db_error": None,
    }

    try:
        backend = detect_db_type(db_url)
    except Exception as exc:
        info["db_error"] = f"detect_db_type_failed: {exc}"
        backend = None

    info["db_backend"] = backend
    if backend != "clickhouse":
        if missing_migrations:
            info["status"] = "missing"
        return info

    try:
        sink = ClickHouseMetricsSink(db_url)
    except Exception as exc:
        info["db_error"] = str(exc)
        if missing_migrations:
            info["status"] = "missing"
        return info

    try:
        client = sink.client
        info["checked"] = True
        table_rows = _query_dicts(
            client,
            """
            SELECT name
            FROM system.tables
            WHERE database = currentDatabase()
              AND name IN %(tables)s
            """,
            {"tables": list(REQUIRED_TABLES)},
        )
        present_tables = {row.get("name") for row in table_rows}
        missing_tables = [t for t in REQUIRED_TABLES if t not in present_tables]
        info["missing_tables"] = missing_tables

        missing_columns: Dict[str, List[str]] = {}
        for table in REQUIRED_TABLES:
            if table in missing_tables:
                continue
            col_rows = _query_dicts(
                client,
                """
                SELECT name
                FROM system.columns
                WHERE database = currentDatabase()
                  AND table = %(table)s
                """,
                {"table": table},
            )
            present_columns = {row.get("name") for row in col_rows}
            required_cols = REQUIRED_COLUMNS.get(table, set())
            missing = sorted(required_cols - present_columns)
            if missing:
                missing_columns[table] = missing
        info["missing_columns"] = missing_columns

        if missing_tables or missing_columns:
            info["status"] = "missing"
        else:
            info["status"] = "ok"
    except Exception as exc:
        info["db_error"] = str(exc)
        if missing_migrations:
            info["status"] = "missing"
    finally:
        sink.close()

    return info


def compile_coverage_report(
    *,
    providers: Sequence[str],
    collector: Dict[str, Dict[str, Any]],
    config: Dict[str, Dict[str, Any]],
    schema: Dict[str, Any],
    sink: Dict[str, Dict[str, Any]],
    commands: Dict[str, Dict[str, Any]],
) -> Dict[str, Any]:
    report: Dict[str, Any] = {
        "providers": {},
        "schema": schema,
        "overall_ok": True,
    }

    schema_status = schema.get("status")
    schema_ok = schema_status != "missing"

    overall_ok = True
    for provider in providers:
        collector_entry = collector.get(provider, {"status": "missing"})
        config_entry = config.get(provider, {"status": "missing"})
        sink_entry = sink.get(provider, {"status": "missing"})
        commands_entry = commands.get(provider, {"status": "missing"})

        provider_ok = (
            collector_entry.get("status") == "ok"
            and config_entry.get("status") == "ok"
            and sink_entry.get("status") == "ok"
            and commands_entry.get("status") == "ok"
            and schema_ok
        )
        overall_ok = overall_ok and provider_ok

        report["providers"][provider] = {
            "collector": collector_entry,
            "config": config_entry,
            "schema": schema,
            "sink": sink_entry,
            "commands": commands_entry,
            "overall": "ok" if provider_ok else "missing",
        }

    report["overall_ok"] = overall_ok
    return report


def run_coverage_audit(*, db_url: str, providers: Sequence[str]) -> Dict[str, Any]:
    repo_root = Path(__file__).resolve().parents[3]
    collector = _collector_check(repo_root)
    config = _config_check(repo_root)
    sink = _sink_check(repo_root)
    commands = _commands_check(repo_root)
    schema = _schema_check(db_url, repo_root)
    return compile_coverage_report(
        providers=providers,
        collector=collector,
        config=config,
        schema=schema,
        sink=sink,
        commands=commands,
    )


def _render_table(headers: List[str], rows: List[List[str]]) -> str:
    widths = [len(h) for h in headers]
    for row in rows:
        for idx, cell in enumerate(row):
            widths[idx] = max(widths[idx], len(cell))

    def _row(cells: List[str]) -> str:
        padded = [cell.ljust(widths[idx]) for idx, cell in enumerate(cells)]
        return f"| {' | '.join(padded)} |"

    lines = [_row(headers)]
    lines.append(_row(["-" * w for w in widths]))
    for row in rows:
        lines.append(_row(row))
    return "\n".join(lines)


def format_coverage_table(report: Dict[str, Any]) -> str:
    providers = report.get("providers", {})
    rows: List[List[str]] = []
    for provider in providers:
        entry = providers.get(provider, {})
        rows.append(
            [
                provider,
                entry.get("collector", {}).get("status", "missing"),
                entry.get("config", {}).get("status", "missing"),
                entry.get("schema", {}).get("status", "missing"),
                entry.get("sink", {}).get("status", "missing"),
                entry.get("commands", {}).get("status", "missing"),
                entry.get("overall", "missing"),
            ]
        )

    return _render_table(
        [
            "provider",
            "collector",
            "config",
            "schema",
            "sink",
            "commands",
            "overall",
        ],
        rows,
    )


def _serialize_report(value: Any) -> Any:
    if isinstance(value, dict):
        return {key: _serialize_report(val) for key, val in value.items()}
    if isinstance(value, list):
        return [_serialize_report(val) for val in value]
    return value


def format_coverage_json(report: Dict[str, Any]) -> str:
    return json.dumps(_serialize_report(report), indent=2, sort_keys=True)


def coverage_failed(report: Dict[str, Any]) -> bool:
    return not bool(report.get("overall_ok"))


def register_commands(audit_subparsers: argparse._SubParsersAction) -> None:
    audit_coverage = audit_subparsers.add_parser(
        "coverage", help="Audit provider implementation coverage."
    )
    audit_coverage.add_argument(
        "--db", required=True, help="Database connection string."
    )
    audit_coverage.add_argument(
        "--provider",
        help="Comma-separated list of providers to audit (default: all).",
    )
    audit_coverage.add_argument(
        "--format", choices=["table", "json"], default="table", help="Output format."
    )
    audit_coverage.set_defaults(func=_cmd_audit_coverage)


def _cmd_audit_coverage(ns: argparse.Namespace) -> int:
    import logging

    logger = logging.getLogger(__name__)
    try:
        providers = parse_provider_list(ns.provider)
        report = run_coverage_audit(db_url=ns.db, providers=providers)
        if ns.format == "json":
            print(format_coverage_json(report))
        else:
            print(format_coverage_table(report))
        return 0 if report.get("overall_ok") else 1
    except Exception as e:
        logger.error(f"Coverage audit failed: {e}")
        return 1



from __future__ import annotations

import re
from datetime import datetime, timezone
from typing import Any, Dict, List, Optional, Sequence, Tuple
import logging

from dev_health_ops.models.work_items import (
    Sprint,
    WorkItem,
    WorkItemDependency,
    WorkItemInteractionEvent,
    WorkItemReopenEvent,
    WorkItemStatusTransition,
)
from dev_health_ops.providers.identity import IdentityResolver
from dev_health_ops.providers.status_mapping import StatusMapping


def _to_utc(dt: Optional[datetime]) -> Optional[datetime]:
    if dt is None:
        return None
    if dt.tzinfo is None:
        return dt.replace(tzinfo=timezone.utc)
    return dt.astimezone(timezone.utc)


def _labels_from_nodes(nodes: Any) -> List[str]:
    labels: List[str] = []
    for node in nodes or []:
        name = (
            (node or {}).get("name")
            if isinstance(node, dict)
            else getattr(node, "name", None)
        )
        if name:
            labels.append(str(name))
    return labels


def _update_transitions_work_item_id(
    transitions: List[WorkItemStatusTransition],
    work_item_id: str,
) -> List[WorkItemStatusTransition]:
    """
    Update work_item_id in a list of transitions.

    Creates new WorkItemStatusTransition instances with the updated work_item_id
    while preserving all other fields.
    """
    return [
        WorkItemStatusTransition(
            work_item_id=work_item_id,
            provider=t.provider,
            occurred_at=t.occurred_at,
            from_status_raw=t.from_status_raw,
            to_status_raw=t.to_status_raw,
            from_status=t.from_status,
            to_status=t.to_status,
            actor=t.actor,
        )
        for t in transitions
    ]


def github_issue_to_work_item(
    *,
    issue: Any,
    repo_full_name: str,
    repo_id: Optional[Any],
    status_mapping: StatusMapping,
    identity: IdentityResolver,
    events: Optional[Sequence[Any]] = None,
    project_status_raw: Optional[str] = None,
) -> Tuple[WorkItem, List[WorkItemStatusTransition]]:
    number = int(getattr(issue, "number", 0) or 0)
    work_item_id = f"gh:{repo_full_name}#{number}"

    title = getattr(issue, "title", "") or ""
    description = getattr(issue, "body", None)
    state = getattr(issue, "state", None)
    created_at = _to_utc(getattr(issue, "created_at", None)) or datetime.now(
        timezone.utc
    )
    updated_at = _to_utc(getattr(issue, "updated_at", None)) or created_at
    closed_at = _to_utc(getattr(issue, "closed_at", None))

    labels = [getattr(lbl, "name", None) for lbl in getattr(issue, "labels", []) or []]
    labels = [str(lbl) for lbl in labels if lbl]

    # If the issue is in a Project with a status field, prefer that as status_raw.
    status_raw = project_status_raw
    normalized_status = status_mapping.normalize_status(
        provider="github",
        status_raw=status_raw,
        labels=() if status_raw else labels,
        state=str(state) if state else None,
    )
    normalized_type = status_mapping.normalize_type(
        provider="github",
        type_raw=None,
        labels=labels,
    )

    assignees: List[str] = []
    for a in getattr(issue, "assignees", []) or []:
        assignees.append(
            identity.resolve(
                provider="github",
                email=getattr(a, "email", None),
                username=getattr(a, "login", None),
                display_name=getattr(a, "name", None),
            )
        )

    reporter_obj = getattr(issue, "user", None)
    reporter = None
    if reporter_obj is not None:
        reporter = identity.resolve(
            provider="github",
            email=getattr(reporter_obj, "email", None),
            username=getattr(reporter_obj, "login", None),
            display_name=getattr(reporter_obj, "name", None),
        )

    url = getattr(issue, "html_url", None) or getattr(issue, "url", None)

    # Best-effort transitions from issue events (label add/remove, closed/reopened).
    transitions: List[WorkItemStatusTransition] = []
    started_at = None
    completed_at = None
    if events:
        # Events are returned newest-first by PyGithub; sort by created_at.
        def _ev_dt(ev: Any) -> datetime:
            return _to_utc(getattr(ev, "created_at", None)) or datetime.min.replace(
                tzinfo=timezone.utc
            )

        prev_status = "unknown"
        for ev in sorted(list(events), key=_ev_dt):
            event_type = str(getattr(ev, "event", "") or "").lower()
            occurred_at = _to_utc(getattr(ev, "created_at", None)) or created_at

            if event_type in {"closed", "reopened"}:
                to_status = "done" if event_type == "closed" else "todo"
                transitions.append(
                    WorkItemStatusTransition(
                        work_item_id=work_item_id,
                        provider="github",
                        occurred_at=occurred_at,
                        from_status_raw=None,
                        to_status_raw=event_type,
                        from_status=prev_status,  # type: ignore[arg-type]
                        to_status=to_status,  # type: ignore[arg-type]
                        actor=None,
                    )
                )
                prev_status = to_status  # type: ignore[assignment]
                continue

            if event_type not in {"labeled", "unlabeled"}:
                continue

            label_obj = getattr(ev, "label", None)
            label_name = getattr(label_obj, "name", None)
            if not label_name:
                continue
            label_name = str(label_name)

            mapped = status_mapping.normalize_status(
                provider="github",
                status_raw=None,
                labels=[label_name] if event_type == "labeled" else (),
                state=None,
            )
            if mapped == "unknown":
                continue

            transitions.append(
                WorkItemStatusTransition(
                    work_item_id=work_item_id,
                    provider="github",
                    occurred_at=occurred_at,
                    from_status_raw=None,
                    to_status_raw=label_name,
                    from_status=prev_status,  # type: ignore[arg-type]
                    to_status=mapped,
                    actor=None,
                )
            )
            prev_status = mapped

        for t in transitions:
            if started_at is None and t.to_status == "in_progress":
                started_at = t.occurred_at
            if completed_at is None and t.to_status in {"done", "canceled"}:
                completed_at = t.occurred_at
                break

    # Fallback: closed_at implies done.
    if completed_at is None and closed_at is not None:
        completed_at = closed_at

    work_item = WorkItem(
        work_item_id=work_item_id,
        provider="github",
        repo_id=repo_id,
        project_key=None,
        # For work tracking metrics, treat the repo as the "project" scope.
        project_id=str(repo_full_name) if repo_full_name else None,
        title=str(title),
        description=str(description) if description else None,
        type=normalized_type,
        status=normalized_status,
        status_raw=str(status_raw) if status_raw else (str(state) if state else None),
        assignees=[a for a in assignees if a and a != "unknown"],
        reporter=reporter if reporter and reporter != "unknown" else None,
        created_at=created_at,
        updated_at=updated_at,
        started_at=started_at,
        completed_at=completed_at,
        closed_at=closed_at,
        labels=labels,
        url=url,
    )

    return work_item, transitions


def github_project_v2_item_to_work_item(
    *,
    item_node: Dict[str, Any],
    project_scope_id: Optional[str] = None,
    status_mapping: StatusMapping,
    identity: IdentityResolver,
) -> Tuple[Optional[WorkItem], List[WorkItemStatusTransition]]:
    """
    Normalize a Projects v2 item node into a WorkItem with status transitions.

    Parses field changes to reconstruct status/phase transition history.
    """
    content = item_node.get("content") or {}
    typename = content.get("__typename")

    # Extract a status, iteration, and estimate values from field values.
    status_raw = None
    iteration_title = None
    iteration_id = None
    estimate = None

    for fv in (item_node.get("fieldValues") or {}).get("nodes") or []:
        fv_typename = (fv or {}).get("__typename")
        field = (fv or {}).get("field") or {}
        field_name = str(field.get("name") or "").strip().lower()

        if fv_typename == "ProjectV2ItemFieldSingleSelectValue":
            if field_name == "status":
                status_raw = fv.get("name")

        elif fv_typename == "ProjectV2ItemFieldIterationValue":
            # GitHub Iterations
            if "iteration" in field_name or "sprint" in field_name:
                iteration_title = fv.get("title")
                iteration_id = fv.get("id")  # internal node id

        elif fv_typename == "ProjectV2ItemFieldNumberValue":
            # Estimates / Points
            if field_name in {"estimate", "points", "story points", "size"}:
                try:
                    estimate = float(fv.get("number") or 0)
                except (ValueError, TypeError):
                    logging.getLogger(__name__).debug(
                        "Failed to parse numeric estimate from value %r",
                        fv.get("number"),
                    )

    # Parse field changes to create status transitions
    transitions: List[WorkItemStatusTransition] = []

    # Extract transitions from changes
    for change in (item_node.get("changes") or {}).get("nodes") or []:
        field = change.get("field") or {}
        field_name = str(field.get("name") or "").strip().lower()

        # Only track status and phase field changes
        if field_name not in {"status", "phase"}:
            continue

        prev_val = (change.get("previousValue") or {}).get("name")
        new_val = (change.get("newValue") or {}).get("name")
        occurred_at = _parse_iso(change.get("createdAt"))

        if not occurred_at or not new_val:
            continue

        # Map to normalized statuses
        from_status = status_mapping.normalize_status(
            provider="github",
            status_raw=str(prev_val) if prev_val else None,
            labels=(),
            state=None,
        )
        to_status = status_mapping.normalize_status(
            provider="github",
            status_raw=str(new_val) if new_val else None,
            labels=(),
            state=None,
        )

        # Get actor
        actor_obj = change.get("actor") or {}
        actor = (
            identity.resolve(
                provider="github",
                email=None,
                username=actor_obj.get("login"),
                display_name=None,
            )
            if actor_obj.get("login")
            else None
        )

        # We'll set work_item_id later when we know the content type
        transitions.append(
            WorkItemStatusTransition(
                work_item_id="",  # Placeholder, will be updated
                provider="github",
                occurred_at=_to_utc(occurred_at) or datetime.now(timezone.utc),
                from_status_raw=str(prev_val) if prev_val else None,
                to_status_raw=str(new_val) if new_val else None,
                from_status=from_status,
                to_status=to_status,
                actor=actor if actor and actor != "unknown" else None,
            )
        )

    if typename == "Issue":
        repo_full_name = ((content.get("repository") or {}).get("nameWithOwner")) or ""
        number = int(content.get("number") or 0)
        work_item_id = (
            f"gh:{repo_full_name}#{number}"
            if repo_full_name and number
            else f"ghproj:{item_node.get('id')}"
        )
        labels = _labels_from_nodes(((content.get("labels") or {}).get("nodes")) or [])
        assignees = []
        for a in ((content.get("assignees") or {}).get("nodes")) or []:
            assignees.append(
                identity.resolve(
                    provider="github",
                    email=a.get("email"),
                    username=a.get("login"),
                    display_name=a.get("name"),
                )
            )
        author = content.get("author") or {}
        reporter = identity.resolve(
            provider="github",
            email=author.get("email"),
            username=author.get("login"),
            display_name=author.get("name"),
        )
        created_at = _to_utc(_parse_iso(content.get("createdAt"))) or datetime.now(
            timezone.utc
        )
        updated_at = _to_utc(_parse_iso(content.get("updatedAt"))) or created_at
        closed_at = _to_utc(_parse_iso(content.get("closedAt")))
        state = content.get("state")

        normalized_status = status_mapping.normalize_status(
            provider="github",
            status_raw=str(status_raw) if status_raw else None,
            labels=() if status_raw else labels,
            state=str(state) if state else None,
        )
        normalized_type = status_mapping.normalize_type(
            provider="github",
            type_raw=None,
            labels=labels,
        )
        completed_at = closed_at if closed_at else None

        # Update work_item_id in transitions
        transitions = _update_transitions_work_item_id(transitions, work_item_id)

        description = content.get("body")
        return WorkItem(
            work_item_id=work_item_id,
            provider="github",
            repo_id=None,
            project_key=None,
            project_id=str(project_scope_id or repo_full_name)
            if (project_scope_id or repo_full_name)
            else None,
            title=str(content.get("title") or ""),
            description=str(description) if description else None,
            type=normalized_type,
            status=normalized_status,
            status_raw=str(status_raw)
            if status_raw
            else (str(state) if state else None),
            assignees=[a for a in assignees if a and a != "unknown"],
            reporter=reporter if reporter and reporter != "unknown" else None,
            created_at=created_at,
            updated_at=updated_at,
            started_at=None,
            completed_at=completed_at,
            closed_at=closed_at,
            labels=labels,
            story_points=estimate,
            sprint_id=iteration_id,
            sprint_name=iteration_title,
            url=content.get("url"),
        ), transitions

    if typename == "DraftIssue":
        created_at = _to_utc(_parse_iso(content.get("createdAt"))) or datetime.now(
            timezone.utc
        )
        updated_at = _to_utc(_parse_iso(content.get("updatedAt"))) or created_at
        normalized_status = status_mapping.normalize_status(
            provider="github",
            status_raw=str(status_raw) if status_raw else None,
            labels=(),
            state=None,
        )

        work_item_id = f"ghproj:{item_node.get('id')}"

        # Update work_item_id in transitions
        transitions = _update_transitions_work_item_id(transitions, work_item_id)

        description = content.get("body")
        return WorkItem(
            work_item_id=work_item_id,
            provider="github",
            repo_id=None,
            project_key=None,
            project_id=str(project_scope_id) if project_scope_id else None,
            title=str(content.get("title") or ""),
            description=str(description) if description else None,
            type="issue",
            status=normalized_status,
            status_raw=str(status_raw) if status_raw else None,
            assignees=[],
            reporter=None,
            created_at=created_at,
            updated_at=updated_at,
            started_at=None,
            completed_at=None,
            closed_at=None,
            labels=[],
            story_points=estimate,
            sprint_id=iteration_id,
            sprint_name=iteration_title,
            url=None,
        ), transitions

    return None, []


def _parse_iso(value: Optional[str]) -> Optional[datetime]:
    if not value:
        return None
    try:
        # GitHub returns RFC3339 with Z.
        return datetime.fromisoformat(str(value).replace("Z", "+00:00"))
    except (ValueError, TypeError):
        return None


def github_pr_to_work_item(
    *,
    pr: Any,
    repo_full_name: str,
    repo_id: Optional[Any],
    status_mapping: StatusMapping,
    identity: IdentityResolver,
    events: Optional[Sequence[Any]] = None,
) -> Tuple[WorkItem, List[WorkItemStatusTransition]]:
    """
    Normalize a GitHub pull request to a WorkItem with status transitions.

    Args:
        pr: PyGithub PullRequest object
        repo_full_name: e.g. "owner/repo"
        repo_id: optional internal repo ID
        status_mapping: StatusMapping instance
        identity: IdentityResolver instance
        events: optional list of PR events (labeled/unlabeled/closed/reopened)

    Returns:
        Tuple of (WorkItem, list of WorkItemStatusTransition)
    """
    number = int(getattr(pr, "number", 0) or 0)
    work_item_id = f"ghpr:{repo_full_name}#{number}"

    title = getattr(pr, "title", "") or ""
    description = getattr(pr, "body", None)
    state = getattr(pr, "state", None)
    merged = getattr(pr, "merged", False)
    created_at = _to_utc(getattr(pr, "created_at", None)) or datetime.now(timezone.utc)
    updated_at = _to_utc(getattr(pr, "updated_at", None)) or created_at
    closed_at = _to_utc(getattr(pr, "closed_at", None))
    merged_at = _to_utc(getattr(pr, "merged_at", None))

    labels = [getattr(lbl, "name", None) for lbl in getattr(pr, "labels", []) or []]
    labels = [str(lbl) for lbl in labels if lbl]

    # Determine status based on state and merge status
    if merged or merged_at:
        status_raw = "merged"
        normalized_status = "done"
    elif state == "closed":
        status_raw = "closed"
        normalized_status = "canceled"
    elif state == "open":
        status_raw = "open"
        # Check labels to see if it's in review, draft, etc.
        draft = getattr(pr, "draft", False)
        if draft:
            normalized_status = "todo"
        else:
            normalized_status = "in_progress"
    else:
        status_raw = str(state) if state else None
        normalized_status = status_mapping.normalize_status(
            provider="github",
            status_raw=status_raw,
            labels=labels,
            state=str(state) if state else None,
        )

    # PRs are always type "pr"
    normalized_type = "pr"

    assignees: List[str] = []
    for assignee in getattr(pr, "assignees", []) or []:
        assignees.append(
            identity.resolve(
                provider="github",
                email=getattr(assignee, "email", None),
                username=getattr(assignee, "login", None),
                display_name=getattr(assignee, "name", None),
            )
        )

    reporter_obj = getattr(pr, "user", None)
    reporter = None
    if reporter_obj is not None:
        reporter = identity.resolve(
            provider="github",
            email=getattr(reporter_obj, "email", None),
            username=getattr(reporter_obj, "login", None),
            display_name=getattr(reporter_obj, "name", None),
        )

    url = getattr(pr, "html_url", None) or getattr(pr, "url", None)

    # Build transitions from events
    transitions: List[WorkItemStatusTransition] = []
    started_at = created_at  # PRs are in progress when opened
    completed_at = None

    if events:

        def _ev_dt(ev: Any) -> datetime:
            return _to_utc(getattr(ev, "created_at", None)) or datetime.min.replace(
                tzinfo=timezone.utc
            )

        prev_status = "in_progress"  # PRs start as in_progress
        # Track if we've seen a merged event to determine closed status correctly
        merged_in_history = False

        for ev in sorted(list(events), key=_ev_dt):
            event_type = str(getattr(ev, "event", "") or "").lower()
            occurred_at = _to_utc(getattr(ev, "created_at", None)) or created_at

            if event_type == "merged":
                merged_in_history = True
                transitions.append(
                    WorkItemStatusTransition(
                        work_item_id=work_item_id,
                        provider="github",
                        occurred_at=occurred_at,
                        from_status_raw=None,
                        to_status_raw="merged",
                        from_status=prev_status,  # type: ignore[arg-type]
                        to_status="done",
                        actor=None,
                    )
                )
                prev_status = "done"
            elif event_type == "closed":
                # Use merged_in_history to determine status at the time of close
                to_status = "done" if merged_in_history else "canceled"
                transitions.append(
                    WorkItemStatusTransition(
                        work_item_id=work_item_id,
                        provider="github",
                        occurred_at=occurred_at,
                        from_status_raw=None,
                        to_status_raw=event_type,
                        from_status=prev_status,  # type: ignore[arg-type]
                        to_status=to_status,  # type: ignore[arg-type]
                        actor=None,
                    )
                )
                prev_status = to_status  # type: ignore[assignment]
            elif event_type == "reopened":
                transitions.append(
                    WorkItemStatusTransition(
                        work_item_id=work_item_id,
                        provider="github",
                        occurred_at=occurred_at,
                        from_status_raw=None,
                        to_status_raw="reopened",
                        from_status=prev_status,  # type: ignore[arg-type]
                        to_status="in_progress",
                        actor=None,
                    )
                )
                prev_status = "in_progress"

    # Set completed_at based on final status
    if merged_at:
        completed_at = merged_at
    elif closed_at and normalized_status in {"done", "canceled"}:
        completed_at = closed_at

    work_item = WorkItem(
        work_item_id=work_item_id,
        provider="github",
        repo_id=repo_id,
        project_key=None,
        project_id=str(repo_full_name) if repo_full_name else None,
        title=str(title),
        description=str(description) if description else None,
        type=normalized_type,
        status=normalized_status,
        status_raw=status_raw,
        assignees=[a for a in assignees if a and a != "unknown"],
        reporter=reporter if reporter and reporter != "unknown" else None,
        created_at=created_at,
        updated_at=updated_at,
        started_at=started_at,
        completed_at=completed_at,
        closed_at=closed_at,
        labels=labels,
        url=url,
    )

    return work_item, transitions


def github_comment_to_interaction_event(
    *,
    comment: Any,
    work_item_id: str,
    identity: IdentityResolver,
) -> Optional[WorkItemInteractionEvent]:
    """
    Convert a GitHub comment to a WorkItemInteractionEvent.

    Args:
        comment: PyGithub IssueComment or PullRequestComment object
        work_item_id: work item identifier
        identity: IdentityResolver instance

    Returns:
        WorkItemInteractionEvent or None if invalid
    """
    comment_id = getattr(comment, "id", None)
    if not comment_id:
        return None

    created_at = _to_utc(getattr(comment, "created_at", None))
    if not created_at:
        return None

    user = getattr(comment, "user", None)
    actor = None
    if user:
        actor = identity.resolve(
            provider="github",
            email=getattr(user, "email", None),
            username=getattr(user, "login", None),
            display_name=getattr(user, "name", None),
        )

    body = getattr(comment, "body", "") or ""

    return WorkItemInteractionEvent(
        work_item_id=work_item_id,
        provider="github",
        interaction_type="comment",
        occurred_at=created_at,
        actor=actor if actor and actor != "unknown" else None,
        body_length=len(body),
        last_synced=datetime.now(timezone.utc),
    )


def github_milestone_to_sprint(
    *,
    milestone: Any,
    repo_full_name: str,
) -> Sprint:
    """
    Convert a GitHub milestone to a Sprint.

    Args:
        milestone: PyGithub Milestone object
        repo_full_name: e.g. "owner/repo"

    Returns:
        Sprint object
    """
    milestone_id = str(getattr(milestone, "id", 0) or getattr(milestone, "number", 0))
    sprint_id = f"ghms:{repo_full_name}:{milestone_id}"

    title = getattr(milestone, "title", "") or ""
    created_at = _to_utc(getattr(milestone, "created_at", None)) or datetime.now(
        timezone.utc
    )
    due_on = _to_utc(getattr(milestone, "due_on", None))
    state = getattr(milestone, "state", "open")

    return Sprint(
        provider="github",
        sprint_id=sprint_id,
        name=title,
        state="closed" if state == "closed" else "active",
        started_at=created_at,
        ended_at=due_on,
        completed_at=due_on if state == "closed" else None,
        last_synced=datetime.now(timezone.utc),
    )


def detect_github_reopen_events(
    *,
    work_item_id: str,
    events: Sequence[Any],
    identity: IdentityResolver,
) -> List[WorkItemReopenEvent]:
    """
    Detect reopen events from GitHub issue/PR events.

    Args:
        work_item_id: work item identifier
        events: list of issue/PR events from GitHub API
        identity: IdentityResolver instance

    Returns:
        List of WorkItemReopenEvent objects
    """
    reopen_events: List[WorkItemReopenEvent] = []

    for ev in events:
        event_type = str(getattr(ev, "event", "") or "").lower()
        if event_type != "reopened":
            continue

        occurred_at = _to_utc(getattr(ev, "created_at", None))
        if not occurred_at:
            continue

        actor_obj = getattr(ev, "actor", None)
        actor = None
        if actor_obj:
            actor = identity.resolve(
                provider="github",
                email=getattr(actor_obj, "email", None),
                username=getattr(actor_obj, "login", None),
                display_name=getattr(actor_obj, "name", None),
            )

        reopen_events.append(
            WorkItemReopenEvent(
                work_item_id=work_item_id,
                occurred_at=occurred_at,
                from_status="done",  # Was closed before reopen
                to_status="todo",  # Reopened to todo
                from_status_raw="closed",
                to_status_raw="reopened",
                actor=actor if actor and actor != "unknown" else None,
                last_synced=datetime.now(timezone.utc),
            )
        )

    return reopen_events


# Regex patterns for parsing GitHub-style references from issue/PR body
_GITHUB_ISSUE_REF_PATTERN = re.compile(
    r"(?:^|[^\S\r\n])(?:depends\s+on|blocked\s+by|blocks|fixes|closes|resolves)\s*:?\s*"
    r"(?:#(\d+)|(?:(?:https?://)?github\.com/)?([a-zA-Z0-9_-]+/[a-zA-Z0-9_.-]+)#(\d+))",
    re.IGNORECASE | re.MULTILINE,
)


def extract_github_dependencies(
    *,
    work_item_id: str,
    issue_or_pr: Any,
    repo_full_name: str,
) -> List[WorkItemDependency]:
    """
    Extract dependencies from GitHub issue/PR body text.

    Parses common patterns like:
    - "depends on #123"
    - "blocked by owner/repo#456"
    - "blocks #789"

    Args:
        work_item_id: source work item identifier
        issue_or_pr: PyGithub Issue or PullRequest object
        repo_full_name: e.g. "owner/repo"

    Returns:
        List of WorkItemDependency objects
    """
    dependencies: List[WorkItemDependency] = []
    body = getattr(issue_or_pr, "body", "") or ""

    for match in _GITHUB_ISSUE_REF_PATTERN.finditer(body):
        keyword = match.group(0).strip().lower()

        # Determine direction and dependency type
        if any(k in keyword for k in ["depends on", "blocked by"]):
            dep_type = "blocks"  # The referenced issue blocks this one
        elif "blocks" in keyword:
            dep_type = "is_blocked_by"  # This issue blocks the referenced one
        elif any(k in keyword for k in ["fixes", "closes", "resolves"]):
            dep_type = "relates_to"
        else:
            dep_type = "relates_to"

        # Parse reference
        same_repo_num = match.group(1)
        cross_repo = match.group(2)
        cross_repo_num = match.group(3)

        if same_repo_num:
            target_id = f"gh:{repo_full_name}#{same_repo_num}"
        elif cross_repo and cross_repo_num:
            target_id = f"gh:{cross_repo}#{cross_repo_num}"
        else:
            continue

        dependencies.append(
            WorkItemDependency(
                source_work_item_id=work_item_id,
                target_work_item_id=target_id,
                relationship_type=dep_type,
                relationship_type_raw=dep_type,
                last_synced=datetime.now(timezone.utc),
            )
        )

    return dependencies


# Priority label mapping - maps label to (priority_raw, service_class)
_PRIORITY_LABELS = {
    "priority::critical": ("critical", "expedite"),
    "priority::high": ("high", "fixed_date"),
    "priority::medium": ("medium", "standard"),
    "priority::low": ("low", "intangible"),
    "p0": ("critical", "expedite"),
    "p1": ("high", "fixed_date"),
    "p2": ("medium", "standard"),
    "p3": ("low", "intangible"),
    "priority-critical": ("critical", "expedite"),
    "priority-high": ("high", "fixed_date"),
    "priority-medium": ("medium", "standard"),
    "priority-low": ("low", "intangible"),
    "critical": ("critical", "expedite"),
    "urgent": ("critical", "expedite"),
    "high-priority": ("high", "fixed_date"),
    "low-priority": ("low", "intangible"),
}


def _priority_from_labels(labels: Sequence[str]) -> Tuple[Optional[str], Optional[str]]:
    """
    Extract priority_raw and service_class from GitHub labels.

    Returns (priority_raw, service_class) or (None, None) if no match.
    """
    for label in labels:
        normalized = label.lower().strip()
        if normalized in _PRIORITY_LABELS:
            return _PRIORITY_LABELS[normalized]
    return (None, None)


def enrich_work_item_with_priority(
    work_item: WorkItem,
    labels: Sequence[str],
) -> WorkItem:
    """
    Enrich a WorkItem with priority and service_class from labels.

    Args:
        work_item: WorkItem to enrich
        labels: list of label strings

    Returns:
        WorkItem with priority/service_class set (or original if no match)
    """
    priority, service_class = _priority_from_labels(labels)
    if priority is None:
        return work_item

    return WorkItem(
        work_item_id=work_item.work_item_id,
        provider=work_item.provider,
        repo_id=work_item.repo_id,
        project_key=work_item.project_key,
        project_id=work_item.project_id,
        title=work_item.title,
        description=work_item.description,
        type=work_item.type,
        status=work_item.status,
        status_raw=work_item.status_raw,
        assignees=work_item.assignees,
        reporter=work_item.reporter,
        created_at=work_item.created_at,
        updated_at=work_item.updated_at,
        started_at=work_item.started_at,
        completed_at=work_item.completed_at,
        closed_at=work_item.closed_at,
        labels=work_item.labels,
        story_points=work_item.story_points,
        sprint_id=work_item.sprint_id,
        sprint_name=work_item.sprint_name,
        priority_raw=priority,
        service_class=service_class,
        url=work_item.url,
    )



import pytest
from unittest.mock import AsyncMock
from dev_health_ops.llm.providers.openai import OpenAIProvider


@pytest.mark.asyncio
async def test_openai_retry_on_empty_content():
    # Legacy models use chat completions
    provider = OpenAIProvider(api_key="test", model="gpt-4o")

    # Mock client and response
    mock_client = AsyncMock()

    # First response is empty, second is valid JSON
    mock_response_empty = AsyncMock()
    mock_response_empty.choices = [
        AsyncMock(message=AsyncMock(content=""), finish_reason="stop")
    ]

    mock_response_valid = AsyncMock()
    mock_response_valid.choices = [
        AsyncMock(
            message=AsyncMock(content='{"summary": "test"}'), finish_reason="stop"
        )
    ]

    mock_client.chat.completions.create.side_effect = [
        mock_response_empty,
        mock_response_valid,
    ]

    # Patch the internal implementation's client
    provider._impl._client = mock_client

    result = await provider.complete("test prompt")

    assert result == '{"summary": "test"}'
    assert mock_client.chat.completions.create.call_count == 2

    # Check that tokens were doubled on retry
    second_call_kwargs = mock_client.chat.completions.create.call_args_list[1].kwargs
    assert second_call_kwargs["max_completion_tokens"] == 8192


@pytest.mark.asyncio
async def test_gpt5_retry_on_finish_reason_truncation():
    # GPT-5 uses responses API and 'max_output_tokens'
    provider = OpenAIProvider(
        api_key="test", model="gpt-5-mini", max_completion_tokens=512
    )

    mock_client = AsyncMock()

    # First response truncated
    mock_response_trunc = AsyncMock()
    mock_response_trunc.output_text = '{"summary": "truncated'
    mock_response_trunc.incomplete_details = AsyncMock(reason="max_output_tokens")

    # Second response valid
    mock_response_valid = AsyncMock()
    mock_response_valid.output_text = '{"summary": "complete"}'
    mock_response_valid.incomplete_details = None

    mock_client.responses.create.side_effect = [
        mock_response_trunc,
        mock_response_valid,
    ]

    provider._impl._client = mock_client

    result = await provider.complete("test prompt")

    assert result == '{"summary": "complete"}'
    assert mock_client.responses.create.call_count == 2

    # Verify token doubling in second call
    second_call_kwargs = mock_client.responses.create.call_args_list[1].kwargs
    assert second_call_kwargs["max_output_tokens"] == 8192


@pytest.mark.asyncio
async def test_openai_token_clamping():
    # Facade clamps tokens to min 1024
    provider = OpenAIProvider(api_key="test", model="gpt-4o", max_completion_tokens=128)

    mock_client = AsyncMock()
    mock_response = AsyncMock()
    mock_response.choices = [
        AsyncMock(message=AsyncMock(content="{}"), finish_reason="stop")
    ]
    mock_client.chat.completions.create.return_value = mock_response

    provider._impl._client = mock_client
    await provider.complete("test")

    # Check that tokens were clamped to min 1024
    call_kwargs = mock_client.chat.completions.create.call_args.kwargs
    assert call_kwargs["max_completion_tokens"] == 4096
